import errno
import json
import os
from random import randbytes
import shutil
import socket
import stat
import time
import unittest

import pytest

from ... import platform
from ...constants import *  # NOQA
from ...manifest import Manifest
from ...platform import is_cygwin, is_win32, is_darwin
from ...repository import Repository
from .. import has_lchflags
from .. import changedir
from .. import (
    are_symlinks_supported,
    are_hardlinks_supported,
    are_fifos_supported,
    is_utime_fully_supported,
    is_birthtime_fully_supported,
    same_ts_ns,
)
from . import (
    ArchiverTestCaseBase,
    ArchiverTestCaseBinaryBase,
    RemoteArchiverTestCaseBase,
    RK_ENCRYPTION,
    BORG_EXES,
    requires_hardlinks,
)


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_basic_functionality(self):
        have_root = self.create_test_files()
        # fork required to test show-rc output
        output = self.cmd(
            f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION, "--show-version", "--show-rc", fork=True
        )
        self.assert_in("borgbackup version", output)
        self.assert_in("terminating with success status, rc 0", output)
        self.cmd(f"--repo={self.repository_location}", "create", "--exclude-nodump", "test", "input")
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "--exclude-nodump", "--stats", "test.2", "input"
        )
        self.assert_in("Archive name: test.2", output)
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test")
        list_output = self.cmd(f"--repo={self.repository_location}", "rlist", "--short")
        self.assert_in("test", list_output)
        self.assert_in("test.2", list_output)
        expected = [
            "input",
            "input/bdev",
            "input/cdev",
            "input/dir2",
            "input/dir2/file2",
            "input/empty",
            "input/file1",
            "input/flagfile",
        ]
        if are_fifos_supported():
            expected.append("input/fifo1")
        if are_symlinks_supported():
            expected.append("input/link1")
        if are_hardlinks_supported():
            expected.append("input/hardlink")
        if not have_root:
            # we could not create these device files without (fake)root
            expected.remove("input/bdev")
            expected.remove("input/cdev")
        if has_lchflags:
            # remove the file we did not backup, so input and output become equal
            expected.remove("input/flagfile")  # this file is UF_NODUMP
            os.remove(os.path.join("input", "flagfile"))
        list_output = self.cmd(f"--repo={self.repository_location}", "list", "test", "--short")
        for name in expected:
            self.assert_in(name, list_output)
        self.assert_dirs_equal("input", "output/input")
        info_output = self.cmd(f"--repo={self.repository_location}", "info", "-a", "test")
        item_count = 5 if has_lchflags else 6  # one file is UF_NODUMP
        self.assert_in("Number of files: %d" % item_count, info_output)
        shutil.rmtree(self.cache_path)
        info_output2 = self.cmd(f"--repo={self.repository_location}", "info", "-a", "test")

        def filter(output):
            # filter for interesting "info" output, ignore cache rebuilding related stuff
            prefixes = ["Name:", "Fingerprint:", "Number of files:", "This archive:", "All archives:", "Chunk index:"]
            result = []
            for line in output.splitlines():
                for prefix in prefixes:
                    if line.startswith(prefix):
                        result.append(line)
            return "\n".join(result)

        # the interesting parts of info_output2 and info_output should be same
        self.assert_equal(filter(info_output), filter(info_output2))

    @requires_hardlinks
    def test_create_duplicate_root(self):
        # setup for #5603
        path_a = os.path.join(self.input_path, "a")
        path_b = os.path.join(self.input_path, "b")
        os.mkdir(path_a)
        os.mkdir(path_b)
        hl_a = os.path.join(path_a, "hardlink")
        hl_b = os.path.join(path_b, "hardlink")
        self.create_regular_file(hl_a, contents=b"123456")
        os.link(hl_a, hl_b)
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=none")
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input", "input")  # give input twice!
        # test if created archive has 'input' contents twice:
        archive_list = self.cmd(f"--repo={self.repository_location}", "list", "test", "--json-lines")
        paths = [json.loads(line)["path"] for line in archive_list.split("\n") if line]
        # we have all fs items exactly once!
        assert sorted(paths) == ["input", "input/a", "input/a/hardlink", "input/b", "input/b/hardlink"]

    @pytest.mark.skipif(is_win32, reason="unix sockets not available on windows")
    def test_unix_socket(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.bind(os.path.join(self.input_path, "unix-socket"))
        except PermissionError as err:
            if err.errno == errno.EPERM:
                pytest.skip("unix sockets disabled or not supported")
            elif err.errno == errno.EACCES:
                pytest.skip("permission denied to create unix sockets")
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        sock.close()
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test")
            assert not os.path.exists("input/unix-socket")

    @pytest.mark.skipif(not is_utime_fully_supported(), reason="cannot properly setup and execute test without utime")
    @pytest.mark.skipif(
        not is_birthtime_fully_supported(), reason="cannot properly setup and execute test without birthtime"
    )
    def test_nobirthtime(self):
        self.create_test_files()
        birthtime, mtime, atime = 946598400, 946684800, 946771200
        os.utime("input/file1", (atime, birthtime))
        os.utime("input/file1", (atime, mtime))
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input", "--nobirthtime")
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test")
        sti = os.stat("input/file1")
        sto = os.stat("output/input/file1")
        assert same_ts_ns(sti.st_birthtime * 1e9, birthtime * 1e9)
        assert same_ts_ns(sto.st_birthtime * 1e9, mtime * 1e9)
        assert same_ts_ns(sti.st_mtime_ns, sto.st_mtime_ns)
        assert same_ts_ns(sto.st_mtime_ns, mtime * 1e9)

    def test_create_stdin(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        input_data = b"\x00foo\n\nbar\n   \n"
        self.cmd(f"--repo={self.repository_location}", "create", "test", "-", input=input_data)
        item = json.loads(self.cmd(f"--repo={self.repository_location}", "list", "test", "--json-lines"))
        assert item["size"] == len(input_data)
        assert item["path"] == "stdin"
        extracted_data = self.cmd(
            f"--repo={self.repository_location}", "extract", "test", "--stdout", binary_output=True
        )
        assert extracted_data == input_data

    def test_create_stdin_checkpointing(self):
        chunk_size = 1000  # fixed chunker with this size, also volume based checkpointing after that volume
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        input_data = b"X" * (chunk_size * 2 - 1)  # one full and one partial chunk
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            f"--chunker-params=fixed,{chunk_size}",
            f"--checkpoint-volume={chunk_size}",
            "test",
            "-",
            input=input_data,
        )
        # repo looking good overall? checks for rc == 0.
        self.cmd(f"--repo={self.repository_location}", "check", "--debug")
        # verify that there are no part files in final archive
        out = self.cmd(f"--repo={self.repository_location}", "list", "test")
        assert "stdin.borg_part" not in out
        # verify full file
        out = self.cmd(f"--repo={self.repository_location}", "extract", "test", "stdin", "--stdout", binary_output=True)
        assert out == input_data

    def test_create_erroneous_file(self):
        chunk_size = 1000  # fixed chunker with this size, also volume based checkpointing after that volume
        self.create_regular_file(os.path.join(self.input_path, "file1"), size=chunk_size * 2)
        self.create_regular_file(os.path.join(self.input_path, "file2"), size=chunk_size * 2)
        self.create_regular_file(os.path.join(self.input_path, "file3"), size=chunk_size * 2)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        flist = "".join(f"input/file{n}\n" for n in range(1, 4))
        out = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            f"--chunker-params=fail,{chunk_size},rrrEEErrrr",
            "--paths-from-stdin",
            "--list",
            "test",
            input=flist.encode(),
            exit_code=0,
        )
        assert "E input/file2" not in out  # we managed to read it in the 3rd retry (after 3 failed reads)
        # repo looking good overall? checks for rc == 0.
        self.cmd(f"--repo={self.repository_location}", "check", "--debug")
        # check files in created archive
        out = self.cmd(f"--repo={self.repository_location}", "list", "test")
        assert "input/file1" in out
        assert "input/file2" in out
        assert "input/file3" in out

    def test_create_content_from_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        input_data = "some test content"
        name = "a/b/c"
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "--stdin-name",
            name,
            "--content-from-command",
            "test",
            "--",
            "echo",
            input_data,
        )
        item = json.loads(self.cmd(f"--repo={self.repository_location}", "list", "test", "--json-lines"))
        assert item["size"] == len(input_data) + 1  # `echo` adds newline
        assert item["path"] == name
        extracted_data = self.cmd(f"--repo={self.repository_location}", "extract", "test", "--stdout")
        assert extracted_data == input_data + "\n"

    def test_create_content_from_command_with_failed_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "--content-from-command",
            "test",
            "--",
            "sh",
            "-c",
            "exit 73;",
            exit_code=2,
        )
        assert output.endswith("Command 'sh' exited with status 73" + os.linesep)
        archive_list = json.loads(self.cmd(f"--repo={self.repository_location}", "rlist", "--json"))
        assert archive_list["archives"] == []

    def test_create_content_from_command_missing_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "test", "--content-from-command", exit_code=2)
        assert output.endswith("No command given." + os.linesep)

    def test_create_paths_from_stdin(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("dir1/file2", size=1024 * 80)
        self.create_regular_file("dir1/file3", size=1024 * 80)
        self.create_regular_file("file4", size=1024 * 80)

        input_data = b"input/file1\0input/dir1\0input/file4"
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "test",
            "--paths-from-stdin",
            "--paths-delimiter",
            "\\0",
            input=input_data,
        )
        archive_list = self.cmd(f"--repo={self.repository_location}", "list", "test", "--json-lines")
        paths = [json.loads(line)["path"] for line in archive_list.split("\n") if line]
        assert paths == ["input/file1", "input/dir1", "input/file4"]

    def test_create_paths_from_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("file2", size=1024 * 80)
        self.create_regular_file("file3", size=1024 * 80)
        self.create_regular_file("file4", size=1024 * 80)

        input_data = "input/file1\ninput/file2\ninput/file3"
        if is_win32:
            with open("filenames.cmd", "w") as script:
                for filename in input_data.splitlines():
                    script.write(f"@echo {filename}\n")
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "--paths-from-command",
            "test",
            "--",
            "filenames.cmd" if is_win32 else "echo",
            input_data,
        )

        archive_list = self.cmd(f"--repo={self.repository_location}", "list", "test", "--json-lines")
        paths = [json.loads(line)["path"] for line in archive_list.split("\n") if line]
        assert paths == ["input/file1", "input/file2", "input/file3"]

    def test_create_paths_from_command_with_failed_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "--paths-from-command",
            "test",
            "--",
            "sh",
            "-c",
            "exit 73;",
            exit_code=2,
        )
        assert output.endswith("Command 'sh' exited with status 73" + os.linesep)
        archive_list = json.loads(self.cmd(f"--repo={self.repository_location}", "rlist", "--json"))
        assert archive_list["archives"] == []

    def test_create_paths_from_command_missing_command(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "test", "--paths-from-command", exit_code=2)
        assert output.endswith("No command given." + os.linesep)

    def test_create_without_root(self):
        """test create without a root"""
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", exit_code=2)

    def test_create_pattern_root(self):
        """test create with only a root pattern"""
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("file2", size=1024 * 80)
        output = self.cmd(f"--repo={self.repository_location}", "create", "test", "-v", "--list", "--pattern=R input")
        self.assert_in("A input/file1", output)
        self.assert_in("A input/file2", output)

    def test_create_pattern(self):
        """test file patterns during create"""
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("file2", size=1024 * 80)
        self.create_regular_file("file_important", size=1024 * 80)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "-v",
            "--list",
            "--pattern=+input/file_important",
            "--pattern=-input/file*",
            "test",
            "input",
        )
        self.assert_in("A input/file_important", output)
        self.assert_in("- input/file1", output)
        self.assert_in("- input/file2", output)

    def test_create_pattern_file(self):
        """test file patterns during create"""
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("file2", size=1024 * 80)
        self.create_regular_file("otherfile", size=1024 * 80)
        self.create_regular_file("file_important", size=1024 * 80)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "-v",
            "--list",
            "--pattern=-input/otherfile",
            "--patterns-from=" + self.patterns_file_path,
            "test",
            "input",
        )
        self.assert_in("A input/file_important", output)
        self.assert_in("- input/file1", output)
        self.assert_in("- input/file2", output)
        self.assert_in("- input/otherfile", output)

    def test_create_pattern_exclude_folder_but_recurse(self):
        """test when patterns exclude a parent folder, but include a child"""
        self.patterns_file_path2 = os.path.join(self.tmpdir, "patterns2")
        with open(self.patterns_file_path2, "wb") as fd:
            fd.write(b"+ input/x/b\n- input/x*\n")

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("x/a/foo_a", size=1024 * 80)
        self.create_regular_file("x/b/foo_b", size=1024 * 80)
        self.create_regular_file("y/foo_y", size=1024 * 80)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "-v",
            "--list",
            "--patterns-from=" + self.patterns_file_path2,
            "test",
            "input",
        )
        self.assert_in("- input/x/a/foo_a", output)
        self.assert_in("A input/x/b/foo_b", output)
        self.assert_in("A input/y/foo_y", output)

    def test_create_pattern_exclude_folder_no_recurse(self):
        """test when patterns exclude a parent folder and, but include a child"""
        self.patterns_file_path2 = os.path.join(self.tmpdir, "patterns2")
        with open(self.patterns_file_path2, "wb") as fd:
            fd.write(b"+ input/x/b\n! input/x*\n")

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("x/a/foo_a", size=1024 * 80)
        self.create_regular_file("x/b/foo_b", size=1024 * 80)
        self.create_regular_file("y/foo_y", size=1024 * 80)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "-v",
            "--list",
            "--patterns-from=" + self.patterns_file_path2,
            "test",
            "input",
        )
        self.assert_not_in("input/x/a/foo_a", output)
        self.assert_not_in("input/x/a", output)
        self.assert_in("A input/y/foo_y", output)

    def test_create_pattern_intermediate_folders_first(self):
        """test that intermediate folders appear first when patterns exclude a parent folder but include a child"""
        self.patterns_file_path2 = os.path.join(self.tmpdir, "patterns2")
        with open(self.patterns_file_path2, "wb") as fd:
            fd.write(b"+ input/x/a\n+ input/x/b\n- input/x*\n")

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)

        self.create_regular_file("x/a/foo_a", size=1024 * 80)
        self.create_regular_file("x/b/foo_b", size=1024 * 80)
        with changedir("input"):
            self.cmd(
                f"--repo={self.repository_location}",
                "create",
                "--patterns-from=" + self.patterns_file_path2,
                "test",
                ".",
            )

        # list the archive and verify that the "intermediate" folders appear before
        # their contents
        out = self.cmd(f"--repo={self.repository_location}", "list", "test", "--format", "{type} {path}{NL}")
        out_list = out.splitlines()

        self.assert_in("d x/a", out_list)
        self.assert_in("d x/b", out_list)

        assert out_list.index("d x/a") < out_list.index("- x/a/foo_a")
        assert out_list.index("d x/b") < out_list.index("- x/b/foo_b")

    def test_create_no_cache_sync(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "rdelete", "--cache-only")
        create_json = json.loads(
            self.cmd(
                f"--repo={self.repository_location}", "create", "--no-cache-sync", "--json", "--error", "test", "input"
            )
        )  # ignore experimental warning
        info_json = json.loads(self.cmd(f"--repo={self.repository_location}", "info", "-a", "test", "--json"))
        create_stats = create_json["cache"]["stats"]
        info_stats = info_json["cache"]["stats"]
        assert create_stats == info_stats
        self.cmd(f"--repo={self.repository_location}", "rdelete", "--cache-only")
        self.cmd(f"--repo={self.repository_location}", "create", "--no-cache-sync", "test2", "input")
        self.cmd(f"--repo={self.repository_location}", "rinfo")
        self.cmd(f"--repo={self.repository_location}", "check")

    def test_create_archivename_with_placeholder(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        ts = "1999-12-31T23:59:59"
        name_given = "test-{now}"  # placeholder in archive name gets replaced by borg
        name_expected = f"test-{ts}"  # placeholder in f-string gets replaced by python
        self.cmd(f"--repo={self.repository_location}", "create", f"--timestamp={ts}", name_given, "input")
        list_output = self.cmd(f"--repo={self.repository_location}", "rlist", "--short")
        assert name_expected in list_output

    def test_exclude_caches(self):
        self._create_test_caches()
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input", "--exclude-caches")
        self._assert_test_caches()

    def test_exclude_tagged(self):
        self._create_test_tagged()
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "test",
            "input",
            "--exclude-if-present",
            ".NOBACKUP",
            "--exclude-if-present",
            "00-NOBACKUP",
        )
        self._assert_test_tagged()

    def test_exclude_keep_tagged(self):
        self._create_test_keep_tagged()
        self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "test",
            "input",
            "--exclude-if-present",
            ".NOBACKUP1",
            "--exclude-if-present",
            ".NOBACKUP2",
            "--exclude-caches",
            "--keep-exclude-tags",
        )
        self._assert_test_keep_tagged()

    def test_path_normalization(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("dir1/dir2/file", size=1024 * 80)
        with changedir("input/dir1/dir2"):
            self.cmd(f"--repo={self.repository_location}", "create", "test", "../../../input/dir1/../dir1/dir2/..")
        output = self.cmd(f"--repo={self.repository_location}", "list", "test")
        self.assert_not_in("..", output)
        self.assert_in(" input/dir1/dir2/file", output)

    def test_exclude_normalization(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("file2", size=1024 * 80)
        with changedir("input"):
            self.cmd(f"--repo={self.repository_location}", "create", "test1", ".", "--exclude=file1")
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test1")
        self.assert_equal(sorted(os.listdir("output")), ["file2"])
        with changedir("input"):
            self.cmd(f"--repo={self.repository_location}", "create", "test2", ".", "--exclude=./file1")
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test2")
        self.assert_equal(sorted(os.listdir("output")), ["file2"])
        self.cmd(f"--repo={self.repository_location}", "create", "test3", "input", "--exclude=input/./file1")
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test3")
        self.assert_equal(sorted(os.listdir("output/input")), ["file2"])

    def test_repeated_files(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input", "input")

    @pytest.mark.skipif("BORG_TESTS_IGNORE_MODES" in os.environ, reason="modes unreliable")
    @pytest.mark.skipif(is_win32, reason="modes unavailable on Windows")
    def test_umask(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        mode = os.stat(self.repository_path).st_mode
        self.assertEqual(stat.S_IMODE(mode), 0o700)

    def test_create_dry_run(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "--dry-run", "test", "input")
        # Make sure no archive has been created
        with Repository(self.repository_path) as repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
        self.assert_equal(len(manifest.archives), 0)

    def test_progress_on(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "test4", "input", "--progress")
        self.assert_in("\r", output)

    def test_progress_off(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "test5", "input")
        self.assert_not_in("\r", output)

    def test_file_status(self):
        """test that various file status show expected results

        clearly incomplete: only tests for the weird "unchanged" status for now"""
        self.create_regular_file("file1", size=1024 * 80)
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "--list", "test", "input")
        self.assert_in("A input/file1", output)
        self.assert_in("A input/file2", output)
        # should find first file as unmodified
        output = self.cmd(f"--repo={self.repository_location}", "create", "--list", "test2", "input")
        self.assert_in("U input/file1", output)
        # this is expected, although surprising, for why, see:
        # https://borgbackup.readthedocs.org/en/latest/faq.html#i-am-seeing-a-added-status-for-a-unchanged-file
        self.assert_in("A input/file2", output)

    @pytest.mark.skipif(
        is_win32, reason="ctime attribute is file creation time on Windows"
    )  # see https://docs.python.org/3/library/os.html#os.stat_result.st_ctime
    def test_file_status_cs_cache_mode(self):
        """test that a changed file with faked "previous" mtime still gets backed up in ctime,size cache_mode"""
        self.create_regular_file("file1", contents=b"123")
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=10)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "test1", "input", "--list", "--files-cache=ctime,size"
        )
        # modify file1, but cheat with the mtime (and atime) and also keep same size:
        st = os.stat("input/file1")
        self.create_regular_file("file1", contents=b"321")
        os.utime("input/file1", ns=(st.st_atime_ns, st.st_mtime_ns))
        # this mode uses ctime for change detection, so it should find file1 as modified
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "test2", "input", "--list", "--files-cache=ctime,size"
        )
        self.assert_in("M input/file1", output)

    def test_file_status_ms_cache_mode(self):
        """test that a chmod'ed file with no content changes does not get chunked again in mtime,size cache_mode"""
        self.create_regular_file("file1", size=10)
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=10)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "--list", "--files-cache=mtime,size", "test1", "input"
        )
        # change mode of file1, no content change:
        st = os.stat("input/file1")
        os.chmod("input/file1", st.st_mode ^ stat.S_IRWXO)  # this triggers a ctime change, but mtime is unchanged
        # this mode uses mtime for change detection, so it should find file1 as unmodified
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "--list", "--files-cache=mtime,size", "test2", "input"
        )
        self.assert_in("U input/file1", output)

    def test_file_status_rc_cache_mode(self):
        """test that files get rechunked unconditionally in rechunk,ctime cache mode"""
        self.create_regular_file("file1", size=10)
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=10)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "--list", "--files-cache=rechunk,ctime", "test1", "input"
        )
        # no changes here, but this mode rechunks unconditionally
        output = self.cmd(
            f"--repo={self.repository_location}", "create", "--list", "--files-cache=rechunk,ctime", "test2", "input"
        )
        self.assert_in("A input/file1", output)

    def test_file_status_excluded(self):
        """test that excluded paths are listed"""

        self.create_regular_file("file1", size=1024 * 80)
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=1024 * 80)
        if has_lchflags:
            self.create_regular_file("file3", size=1024 * 80)
            platform.set_flags(os.path.join(self.input_path, "file3"), stat.UF_NODUMP)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        output = self.cmd(f"--repo={self.repository_location}", "create", "--list", "--exclude-nodump", "test", "input")
        self.assert_in("A input/file1", output)
        self.assert_in("A input/file2", output)
        if has_lchflags:
            self.assert_in("- input/file3", output)
        # should find second file as excluded
        output = self.cmd(
            f"--repo={self.repository_location}",
            "create",
            "test1",
            "input",
            "--list",
            "--exclude-nodump",
            "--exclude",
            "*/file2",
        )
        self.assert_in("U input/file1", output)
        self.assert_in("- input/file2", output)
        if has_lchflags:
            self.assert_in("- input/file3", output)

    def test_file_status_counters(self):
        """Test file status counters in the stats of `borg create --stats`"""

        def to_dict(borg_create_output):
            borg_create_output = borg_create_output.strip().splitlines()
            borg_create_output = [line.split(":", 1) for line in borg_create_output]
            borg_create_output = {
                key: int(value)
                for key, value in borg_create_output
                if key in ("Added files", "Unchanged files", "Modified files")
            }
            return borg_create_output

        # Test case set up: create a repository
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        # Archive an empty dir
        result = self.cmd(f"--repo={self.repository_location}", "create", "--stats", "test_archive", self.input_path)
        result = to_dict(result)
        assert result["Added files"] == 0
        assert result["Unchanged files"] == 0
        assert result["Modified files"] == 0
        # Archive a dir with two added files
        self.create_regular_file("testfile1", contents=b"test1")
        time.sleep(1.0 if is_darwin else 0.01)  # testfile2 must have newer timestamps than testfile1
        self.create_regular_file("testfile2", contents=b"test2")
        result = self.cmd(f"--repo={self.repository_location}", "create", "--stats", "test_archive2", self.input_path)
        result = to_dict(result)
        assert result["Added files"] == 2
        assert result["Unchanged files"] == 0
        assert result["Modified files"] == 0
        # Archive a dir with 1 unmodified file and 1 modified
        self.create_regular_file("testfile1", contents=b"new data")
        result = self.cmd(f"--repo={self.repository_location}", "create", "--stats", "test_archive3", self.input_path)
        result = to_dict(result)
        # Should process testfile2 as added because of
        # https://borgbackup.readthedocs.io/en/stable/faq.html#i-am-seeing-a-added-status-for-an-unchanged-file
        assert result["Added files"] == 1
        assert result["Unchanged files"] == 0
        assert result["Modified files"] == 1

    def test_create_json(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        create_info = json.loads(self.cmd(f"--repo={self.repository_location}", "create", "--json", "test", "input"))
        # The usual keys
        assert "encryption" in create_info
        assert "repository" in create_info
        assert "cache" in create_info
        assert "last_modified" in create_info["repository"]

        archive = create_info["archive"]
        assert archive["name"] == "test"
        assert isinstance(archive["command_line"], str)
        assert isinstance(archive["duration"], float)
        assert len(archive["id"]) == 64
        assert "stats" in archive

    def test_create_topical(self):
        self.create_regular_file("file1", size=1024 * 80)
        time.sleep(1)  # file2 must have newer timestamps than file1
        self.create_regular_file("file2", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        # no listing by default
        output = self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.assert_not_in("file1", output)
        # shouldn't be listed even if unchanged
        output = self.cmd(f"--repo={self.repository_location}", "create", "test0", "input")
        self.assert_not_in("file1", output)
        # should list the file as unchanged
        output = self.cmd(f"--repo={self.repository_location}", "create", "test1", "input", "--list", "--filter=U")
        self.assert_in("file1", output)
        # should *not* list the file as changed
        output = self.cmd(f"--repo={self.repository_location}", "create", "test2", "input", "--list", "--filter=AM")
        self.assert_not_in("file1", output)
        # change the file
        self.create_regular_file("file1", size=1024 * 100)
        # should list the file as changed
        output = self.cmd(f"--repo={self.repository_location}", "create", "test3", "input", "--list", "--filter=AM")
        self.assert_in("file1", output)

    @pytest.mark.skipif(not are_fifos_supported() or is_cygwin, reason="FIFOs not supported, hangs on cygwin")
    def test_create_read_special_symlink(self):
        from threading import Thread

        def fifo_feeder(fifo_fn, data):
            fd = os.open(fifo_fn, os.O_WRONLY)
            try:
                os.write(fd, data)
            finally:
                os.close(fd)

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        data = b"foobar" * 1000

        fifo_fn = os.path.join(self.input_path, "fifo")
        link_fn = os.path.join(self.input_path, "link_fifo")
        os.mkfifo(fifo_fn)
        os.symlink(fifo_fn, link_fn)

        t = Thread(target=fifo_feeder, args=(fifo_fn, data))
        t.start()
        try:
            self.cmd(f"--repo={self.repository_location}", "create", "--read-special", "test", "input/link_fifo")
        finally:
            # In case `borg create` failed to open FIFO, read all data to avoid join() hanging.
            fd = os.open(fifo_fn, os.O_RDONLY | os.O_NONBLOCK)
            try:
                os.read(fd, len(data))
            except OSError:
                # fails on FreeBSD 13 with BlockingIOError
                pass
            finally:
                os.close(fd)
            t.join()
        with changedir("output"):
            self.cmd(f"--repo={self.repository_location}", "extract", "test")
            fifo_fn = "input/link_fifo"
            with open(fifo_fn, "rb") as f:
                extracted_data = f.read()
        assert extracted_data == data

    @pytest.mark.skipif(not are_symlinks_supported(), reason="symlinks not supported")
    def test_create_read_special_broken_symlink(self):
        os.symlink("somewhere does not exist", os.path.join(self.input_path, "link"))
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "--read-special", "test", "input")
        output = self.cmd(f"--repo={self.repository_location}", "list", "test")
        assert "input/link -> somewhere does not exist" in output

    def test_log_json(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        log = self.cmd(
            f"--repo={self.repository_location}", "create", "test", "input", "--log-json", "--list", "--debug"
        )
        messages = {}  # type -> message, one of each kind
        for line in log.splitlines():
            msg = json.loads(line)
            messages[msg["type"]] = msg

        file_status = messages["file_status"]
        assert "status" in file_status
        assert file_status["path"].startswith("input")

        log_message = messages["log_message"]
        assert isinstance(log_message["time"], float)
        assert log_message["levelname"] == "DEBUG"  # there should only be DEBUG messages
        assert isinstance(log_message["message"], str)

    def test_common_options(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        log = self.cmd(f"--repo={self.repository_location}", "--debug", "create", "test", "input")
        assert "security: read previous location" in log

    def test_hashing_time(self):
        def extract_hashing_time(borg_create_output):
            borg_create_output = borg_create_output.strip().splitlines()
            borg_create_output = [line.split(":", 1) for line in borg_create_output]
            hashing_time = [line for line in borg_create_output if line[0] == "Time spent in hashing"].pop()
            hashing_time = hashing_time[1]
            hashing_time = float(hashing_time.removesuffix(" seconds"))
            return hashing_time

        # Test case set up: create a repository and a file
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=none")
        self.create_regular_file("testfile", contents=randbytes(50000000))
        # Archive
        result = self.cmd(f"--repo={self.repository_location}", "create", "--stats", "test_archive", self.input_path)
        hashing_time = extract_hashing_time(result)

        assert hashing_time > 0.0

    def test_chunking_time(self):
        def extract_chunking_time(borg_create_output):
            borg_create_output = borg_create_output.strip().splitlines()
            borg_create_output = [line.split(":", 1) for line in borg_create_output]
            chunking_time = [line for line in borg_create_output if line[0] == "Time spent in chunking"].pop()
            chunking_time = chunking_time[1]
            chunking_time = float(chunking_time.removesuffix(" seconds"))
            return chunking_time

        # Test case set up: create a repository and a file
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("testfile", contents=randbytes(50000000))
        # Archive
        result = self.cmd(f"--repo={self.repository_location}", "create", "--stats", "test_archive", self.input_path)
        chunking_time = extract_chunking_time(result)

        assert chunking_time > 0.0


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    """run the same tests, but with a remote repository"""


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverTestCase):
    """runs the same tests, but via the borg binary"""

    @unittest.skip("test_basic_functionality seems incompatible with fakeroot and/or the binary.")
    def test_basic_functionality(self):
        pass



import argparse
import textwrap
from datetime import timedelta

from ._common import with_repository
from ..archive import Archive
from ..constants import *  # NOQA
from ..helpers import format_timedelta, json_print, basic_json_data
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


class InfoMixIn:
    @with_repository(cache=True, compatibility=(Manifest.Operation.READ,))
    def do_info(self, args, repository, manifest, cache):
        """Show archive details such as disk space used"""

        args.consider_checkpoints = True
        archive_names = tuple(x.name for x in manifest.archives.list_considering(args))

        output_data = []

        for i, archive_name in enumerate(archive_names, 1):
            archive = Archive(manifest, archive_name, cache=cache, iec=args.iec)
            info = archive.info()
            if args.json:
                output_data.append(info)
            else:
                info["duration"] = format_timedelta(timedelta(seconds=info["duration"]))
                print(
                    textwrap.dedent(
                        """
                Archive name: {name}
                Archive fingerprint: {id}
                Comment: {comment}
                Hostname: {hostname}
                Username: {username}
                Time (start): {start}
                Time (end): {end}
                Duration: {duration}
                Command line: {command_line}
                Number of files: {stats[nfiles]}
                Original size: {stats[original_size]}
                Deduplicated size: {stats[deduplicated_size]}
                """
                    )
                    .strip()
                    .format(**info)
                )
            if self.exit_code:
                break
            if not args.json and len(archive_names) - i:
                print()

        if args.json:
            json_print(basic_json_data(manifest, cache=cache, extra={"archives": output_data}))
        return self.exit_code

    def build_parser_info(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog, define_archive_filters_group

        info_epilog = process_epilog(
            """
        This command displays detailed information about the specified archive.

        Please note that the deduplicated sizes of the individual archives do not add
        up to the deduplicated size of the repository ("all archives"), because the two
        are meaning different things:

        This archive / deduplicated size = amount of data stored ONLY for this archive
        = unique chunks of this archive.
        All archives / deduplicated size = amount of data stored in the repo
        = all chunks in the repository.
        """
        )
        subparser = subparsers.add_parser(
            "info",
            parents=[common_parser],
            add_help=False,
            description=self.do_info.__doc__,
            epilog=info_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="show repository or archive information",
        )
        subparser.set_defaults(func=self.do_info)
        subparser.add_argument("--json", action="store_true", help="format output as JSON")
        define_archive_filters_group(subparser)



from tempfile import TemporaryFile

import pytest

from ..helpers.lrucache import LRUCache


class TestLRUCache:
    def test_lrucache(self):
        c = LRUCache(2, dispose=lambda _: None)
        assert len(c) == 0
        assert c.items() == set()
        for i, x in enumerate("abc"):
            c[x] = i
        assert len(c) == 2
        assert c.items() == {("b", 1), ("c", 2)}
        assert "a" not in c
        assert "b" in c
        with pytest.raises(KeyError):
            c["a"]
        assert c.get("a") is None
        assert c.get("a", "foo") == "foo"
        assert c["b"] == 1
        assert c.get("b") == 1
        assert c["c"] == 2
        c["d"] = 3
        assert len(c) == 2
        assert c["c"] == 2
        assert c["d"] == 3
        del c["c"]
        assert len(c) == 1
        with pytest.raises(KeyError):
            c["c"]
        assert c["d"] == 3
        c.clear()
        assert c.items() == set()

    def test_dispose(self):
        c = LRUCache(2, dispose=lambda f: f.close())
        f1 = TemporaryFile()
        f2 = TemporaryFile()
        f3 = TemporaryFile()
        c[1] = f1
        c[2] = f2
        assert not f2.closed
        c[3] = f3
        assert 1 not in c
        assert f1.closed
        assert 2 in c
        assert not f2.closed
        del c[2]
        assert 2 not in c
        assert f2.closed
        c.clear()
        assert c.items() == set()
        assert f3.closed



import argparse

from ._common import with_repository, with_other_repository
from ..cache import Cache
from ..constants import *  # NOQA
from ..crypto.key import key_creator, key_argument_names, tam_required_file
from ..helpers import EXIT_WARNING
from ..helpers import location_validator, Location
from ..helpers import parse_storage_quota
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


class RCreateMixIn:
    @with_repository(create=True, exclusive=True, manifest=False)
    @with_other_repository(manifest=True, compatibility=(Manifest.Operation.READ,))
    def do_rcreate(self, args, repository, *, other_repository=None, other_manifest=None):
        """Create a new, empty repository"""
        other_key = other_manifest.key if other_manifest is not None else None
        path = args.location.canonical_path()
        logger.info('Initializing repository at "%s"' % path)
        if other_key is not None:
            other_key.copy_crypt_key = args.copy_crypt_key
        try:
            key = key_creator(repository, args, other_key=other_key)
        except (EOFError, KeyboardInterrupt):
            repository.destroy()
            return EXIT_WARNING
        manifest = Manifest(key, repository)
        manifest.key = key
        manifest.write()
        repository.commit(compact=False)
        with Cache(repository, manifest, warn_if_unencrypted=False):
            pass
        if key.tam_required:
            tam_file = tam_required_file(repository)
            open(tam_file, "w").close()

        if key.NAME != "plaintext":
            logger.warning(
                "\n"
                "IMPORTANT: you will need both KEY AND PASSPHRASE to access this repo!\n"
                "If you used a repokey mode, the key is stored in the repo, but you should back it up separately.\n"
                'Use "borg key export" to export the key, optionally in printable format.\n'
                "Write down the passphrase. Store both at safe place(s).\n"
            )
        return self.exit_code

    def build_parser_rcreate(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        rcreate_epilog = process_epilog(
            """
        This command creates a new, empty repository. A repository is a filesystem
        directory containing the deduplicated data from zero or more archives.

        Encryption mode TLDR
        ++++++++++++++++++++

        The encryption mode can only be configured when creating a new repository - you can
        neither configure it on a per-archive basis nor change the mode of an existing repository.
        This example will likely NOT give optimum performance on your machine (performance
        tips will come below):

        ::

            borg rcreate --encryption repokey-aes-ocb

        Borg will:

        1. Ask you to come up with a passphrase.
        2. Create a borg key (which contains some random secrets. See :ref:`key_files`).
        3. Derive a "key encryption key" from your passphrase
        4. Encrypt and sign the key with the key encryption key
        5. Store the encrypted borg key inside the repository directory (in the repo config).
           This is why it is essential to use a secure passphrase.
        6. Encrypt and sign your backups to prevent anyone from reading or forging them unless they
           have the key and know the passphrase. Make sure to keep a backup of
           your key **outside** the repository - do not lock yourself out by
           "leaving your keys inside your car" (see :ref:`borg_key_export`).
           The encryption is done locally - if you use a remote repository, the remote machine
           never sees your passphrase, your unencrypted key or your unencrypted files.
           Chunking and id generation are also based on your key to improve
           your privacy.
        7. Use the key when extracting files to decrypt them and to verify that the contents of
           the backups have not been accidentally or maliciously altered.

        Picking a passphrase
        ++++++++++++++++++++

        Make sure you use a good passphrase. Not too short, not too simple. The real
        encryption / decryption key is encrypted with / locked by your passphrase.
        If an attacker gets your key, he can't unlock and use it without knowing the
        passphrase.

        Be careful with special or non-ascii characters in your passphrase:

        - Borg processes the passphrase as unicode (and encodes it as utf-8),
          so it does not have problems dealing with even the strangest characters.
        - BUT: that does not necessarily apply to your OS / VM / keyboard configuration.

        So better use a long passphrase made from simple ascii chars than one that
        includes non-ascii stuff or characters that are hard/impossible to enter on
        a different keyboard layout.

        You can change your passphrase for existing repos at any time, it won't affect
        the encryption/decryption key or other secrets.

        Choosing an encryption mode
        +++++++++++++++++++++++++++

        Depending on your hardware, hashing and crypto performance may vary widely.
        The easiest way to find out about what's fastest is to run ``borg benchmark cpu``.

        `repokey` modes: if you want ease-of-use and "passphrase" security is good enough -
        the key will be stored in the repository (in ``repo_dir/config``).

        `keyfile` modes: if you want "passphrase and having-the-key" security -
        the key will be stored in your home directory (in ``~/.config/borg/keys``).

        The following table is roughly sorted in order of preference, the better ones are
        in the upper part of the table, in the lower part is the old and/or unsafe(r) stuff:

        .. nanorst: inline-fill

        +-----------------------------------+--------------+----------------+--------------------+
        | Mode (K = keyfile or repokey)     | ID-Hash      | Encryption     | Authentication     |
        +-----------------------------------+--------------+----------------+--------------------+
        | K-blake2-chacha20-poly1305        | BLAKE2b      | CHACHA20       | POLY1305           |
        +-----------------------------------+--------------+----------------+--------------------+
        | K-chacha20-poly1305               | HMAC-SHA-256 | CHACHA20       | POLY1305           |
        +-----------------------------------+--------------+----------------+--------------------+
        | K-blake2-aes-ocb                  | BLAKE2b      | AES256-OCB     | AES256-OCB         |
        +-----------------------------------+--------------+----------------+--------------------+
        | K-aes-ocb                         | HMAC-SHA-256 | AES256-OCB     | AES256-OCB         |
        +-----------------------------------+--------------+----------------+--------------------+
        | authenticated-blake2              | BLAKE2b      | none           | BLAKE2b            |
        +-----------------------------------+--------------+----------------+--------------------+
        | authenticated                     | HMAC-SHA-256 | none           | HMAC-SHA256        |
        +-----------------------------------+--------------+----------------+--------------------+
        | none                              | SHA-256      | none           | none               |
        +-----------------------------------+--------------+----------------+--------------------+

        .. nanorst: inline-replace

        `none` mode uses no encryption and no authentication. You're advised NOT to use this mode
        as it would expose you to all sorts of issues (DoS, confidentiality, tampering, ...) in
        case of malicious activity in the repository.

        If you do **not** want to encrypt the contents of your backups, but still want to detect
        malicious tampering use an `authenticated` mode. It's like `repokey` minus encryption.

        Creating a related repository
        +++++++++++++++++++++++++++++

        A related repository uses same secret key material as the other/original repository.

        By default, only the ID key and chunker secret will be the same (these are important
        for deduplication) and the AE crypto keys will be newly generated random keys.

        Optionally, if you use ``--copy-crypt-key`` you can also keep the same crypt_key
        (used for authenticated encryption). Might be desired e.g. if you want to have less
        keys to manage.

        Creating related repositories is useful e.g. if you want to use ``borg transfer`` later.
        """
        )
        subparser = subparsers.add_parser(
            "rcreate",
            parents=[common_parser],
            add_help=False,
            description=self.do_rcreate.__doc__,
            epilog=rcreate_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="create a new, empty repository",
        )
        subparser.set_defaults(func=self.do_rcreate)
        subparser.add_argument(
            "--other-repo",
            metavar="SRC_REPOSITORY",
            dest="other_location",
            type=location_validator(other=True),
            default=Location(other=True),
            help="reuse the key material from the other repository",
        )
        subparser.add_argument(
            "-e",
            "--encryption",
            metavar="MODE",
            dest="encryption",
            required=True,
            choices=key_argument_names(),
            help="select encryption key mode **(required)**",
        )
        subparser.add_argument(
            "--append-only",
            dest="append_only",
            action="store_true",
            help="create an append-only mode repository. Note that this only affects "
            "the low level structure of the repository, and running `delete` "
            "or `prune` will still be allowed. See :ref:`append_only_mode` in "
            "Additional Notes for more details.",
        )
        subparser.add_argument(
            "--storage-quota",
            metavar="QUOTA",
            dest="storage_quota",
            default=None,
            type=parse_storage_quota,
            help="Set storage quota of the new repository (e.g. 5G, 1.5T). Default: no quota.",
        )
        subparser.add_argument(
            "--make-parent-dirs",
            dest="make_parent_dirs",
            action="store_true",
            help="create the parent directories of the repository directory, if they are missing.",
        )
        subparser.add_argument(
            "--copy-crypt-key",
            dest="copy_crypt_key",
            action="store_true",
            help="copy the crypt_key (used for authenticated encryption) from the key of the other repo "
            "(default: new random key).",
        )



from ...constants import *  # NOQA
from . import ArchiverTestCaseBase, RK_ENCRYPTION, environment_variable


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_benchmark_crud(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        with environment_variable(_BORG_BENCHMARK_CRUD_TEST="YES"):
            self.cmd(f"--repo={self.repository_location}", "benchmark", "crud", self.input_path)



import argparse
from collections import OrderedDict
from datetime import datetime, timezone, timedelta
import logging
from operator import attrgetter
import re

from ._common import with_repository
from ..archive import Archive, Statistics
from ..cache import Cache
from ..constants import *  # NOQA
from ..helpers import format_archive, interval, sig_int, log_multi, ProgressIndicatorPercent
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


def prune_within(archives, hours, kept_because):
    target = datetime.now(timezone.utc) - timedelta(seconds=hours * 3600)
    kept_counter = 0
    result = []
    for a in archives:
        if a.ts > target:
            kept_counter += 1
            kept_because[a.id] = ("within", kept_counter)
            result.append(a)
    return result


PRUNING_PATTERNS = OrderedDict(
    [
        ("secondly", "%Y-%m-%d %H:%M:%S"),
        ("minutely", "%Y-%m-%d %H:%M"),
        ("hourly", "%Y-%m-%d %H"),
        ("daily", "%Y-%m-%d"),
        ("weekly", "%G-%V"),
        ("monthly", "%Y-%m"),
        ("yearly", "%Y"),
    ]
)


def prune_split(archives, rule, n, kept_because=None):
    last = None
    keep = []
    pattern = PRUNING_PATTERNS[rule]
    if kept_because is None:
        kept_because = {}
    if n == 0:
        return keep

    a = None
    for a in sorted(archives, key=attrgetter("ts"), reverse=True):
        # we compute the pruning in local time zone
        period = a.ts.astimezone().strftime(pattern)
        if period != last:
            last = period
            if a.id not in kept_because:
                keep.append(a)
                kept_because[a.id] = (rule, len(keep))
                if len(keep) == n:
                    break
    # Keep oldest archive if we didn't reach the target retention count
    if a is not None and len(keep) < n and a.id not in kept_because:
        keep.append(a)
        kept_because[a.id] = (rule + "[oldest]", len(keep))
    return keep


class PruneMixIn:
    @with_repository(exclusive=True, compatibility=(Manifest.Operation.DELETE,))
    def do_prune(self, args, repository, manifest):
        """Prune repository archives according to specified rules"""
        if not any(
            (args.secondly, args.minutely, args.hourly, args.daily, args.weekly, args.monthly, args.yearly, args.within)
        ):
            self.print_error(
                'At least one of the "keep-within", "keep-last", '
                '"keep-secondly", "keep-minutely", "keep-hourly", "keep-daily", '
                '"keep-weekly", "keep-monthly" or "keep-yearly" settings must be specified.'
            )
            return self.exit_code
        checkpoint_re = r"\.checkpoint(\.\d+)?"
        archives_checkpoints = manifest.archives.list(
            match=args.match_archives,
            consider_checkpoints=True,
            match_end=r"(%s)?\Z" % checkpoint_re,
            sort_by=["ts"],
            reverse=True,
        )
        is_checkpoint = re.compile(r"(%s)\Z" % checkpoint_re).search
        checkpoints = [arch for arch in archives_checkpoints if is_checkpoint(arch.name)]
        # keep the latest checkpoint, if there is no later non-checkpoint archive
        if archives_checkpoints and checkpoints and archives_checkpoints[0] is checkpoints[0]:
            keep_checkpoints = checkpoints[:1]
        else:
            keep_checkpoints = []
        checkpoints = set(checkpoints)
        # ignore all checkpoint archives to avoid keeping one (which is an incomplete backup)
        # that is newer than a successfully completed backup - and killing the successful backup.
        archives = [arch for arch in archives_checkpoints if arch not in checkpoints]
        keep = []
        # collect the rule responsible for the keeping of each archive in this dict
        # keys are archive ids, values are a tuple
        #   (<rulename>, <how many archives were kept by this rule so far >)
        kept_because = {}

        # find archives which need to be kept because of the keep-within rule
        if args.within:
            keep += prune_within(archives, args.within, kept_because)

        # find archives which need to be kept because of the various time period rules
        for rule in PRUNING_PATTERNS.keys():
            num = getattr(args, rule, None)
            if num is not None:
                keep += prune_split(archives, rule, num, kept_because)

        to_delete = (set(archives) | checkpoints) - (set(keep) | set(keep_checkpoints))
        stats = Statistics(iec=args.iec)
        with Cache(repository, manifest, lock_wait=self.lock_wait, iec=args.iec) as cache:

            def checkpoint_func():
                manifest.write()
                repository.commit(compact=False)
                cache.commit()

            list_logger = logging.getLogger("borg.output.list")
            # set up counters for the progress display
            to_delete_len = len(to_delete)
            archives_deleted = 0
            uncommitted_deletes = 0
            pi = ProgressIndicatorPercent(total=len(to_delete), msg="Pruning archives %3.0f%%", msgid="prune")
            for archive in archives_checkpoints:
                if sig_int and sig_int.action_done():
                    break
                if archive in to_delete:
                    pi.show()
                    if args.dry_run:
                        log_message = "Would prune:"
                    else:
                        archives_deleted += 1
                        log_message = "Pruning archive (%d/%d):" % (archives_deleted, to_delete_len)
                        archive = Archive(manifest, archive.name, cache)
                        archive.delete(stats, forced=args.forced)
                        checkpointed = self.maybe_checkpoint(
                            checkpoint_func=checkpoint_func, checkpoint_interval=args.checkpoint_interval
                        )
                        uncommitted_deletes = 0 if checkpointed else (uncommitted_deletes + 1)
                else:
                    if is_checkpoint(archive.name):
                        log_message = "Keeping checkpoint archive:"
                    else:
                        log_message = "Keeping archive (rule: {rule} #{num}):".format(
                            rule=kept_because[archive.id][0], num=kept_because[archive.id][1]
                        )
                if args.output_list:
                    list_logger.info(
                        "{message:<40} {archive}".format(message=log_message, archive=format_archive(archive))
                    )
            pi.finish()
            if sig_int:
                # Ctrl-C / SIGINT: do not checkpoint (commit) again, we already have a checkpoint in this case.
                self.print_error("Got Ctrl-C / SIGINT.")
            elif uncommitted_deletes > 0:
                checkpoint_func()
            if args.stats:
                log_multi(str(stats), logger=logging.getLogger("borg.output.stats"))
        return self.exit_code

    def build_parser_prune(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog
        from ._common import define_archive_filters_group

        prune_epilog = process_epilog(
            """
        The prune command prunes a repository by deleting all archives not matching
        any of the specified retention options.

        Important: Repository disk space is **not** freed until you run ``borg compact``.

        This command is normally used by automated backup scripts wanting to keep a
        certain number of historic backups. This retention policy is commonly referred to as
        `GFS <https://en.wikipedia.org/wiki/Backup_rotation_scheme#Grandfather-father-son>`_
        (Grandfather-father-son) backup rotation scheme.

        Also, prune automatically removes checkpoint archives (incomplete archives left
        behind by interrupted backup runs) except if the checkpoint is the latest
        archive (and thus still needed). Checkpoint archives are not considered when
        comparing archive counts against the retention limits (``--keep-X``).

        If you use --match-archives (-a), then only archives that match the pattern are
        considered for deletion and only those archives count towards the totals
        specified by the rules.
        Otherwise, *all* archives in the repository are candidates for deletion!
        There is no automatic distinction between archives representing different
        contents. These need to be distinguished by specifying matching globs.

        If you have multiple sequences of archives with different data sets (e.g.
        from different machines) in one shared repository, use one prune call per
        data set that matches only the respective archives using the --match-archives
        (-a) option.

        The ``--keep-within`` option takes an argument of the form "<int><char>",
        where char is "H", "d", "w", "m", "y". For example, ``--keep-within 2d`` means
        to keep all archives that were created within the past 48 hours.
        "1m" is taken to mean "31d". The archives kept with this option do not
        count towards the totals specified by any other options.

        A good procedure is to thin out more and more the older your backups get.
        As an example, ``--keep-daily 7`` means to keep the latest backup on each day,
        up to 7 most recent days with backups (days without backups do not count).
        The rules are applied from secondly to yearly, and backups selected by previous
        rules do not count towards those of later rules. The time that each backup
        starts is used for pruning purposes. Dates and times are interpreted in the local
        timezone of the system where borg prune runs, and weeks go from Monday to Sunday.
        Specifying a negative number of archives to keep means that there is no limit.
        As of borg 1.2.0, borg will retain the oldest archive if any of the secondly,
        minutely, hourly, daily, weekly, monthly, or yearly rules was not otherwise able to
        meet its retention target. This enables the first chronological archive to continue
        aging until it is replaced by a newer archive that meets the retention criteria.

        The ``--keep-last N`` option is doing the same as ``--keep-secondly N`` (and it will
        keep the last N archives under the assumption that you do not create more than one
        backup archive in the same second).

        When using ``--stats``, you will get some statistics about how much data was
        deleted - the "Deleted data" deduplicated size there is most interesting as
        that is how much your repository will shrink.
        Please note that the "All archives" stats refer to the state after pruning.
        """
        )
        subparser = subparsers.add_parser(
            "prune",
            parents=[common_parser],
            add_help=False,
            description=self.do_prune.__doc__,
            epilog=prune_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="prune archives",
        )
        subparser.set_defaults(func=self.do_prune)
        subparser.add_argument("-n", "--dry-run", dest="dry_run", action="store_true", help="do not change repository")
        subparser.add_argument(
            "--force",
            dest="forced",
            action="store_true",
            help="force pruning of corrupted archives, " "use ``--force --force`` in case ``--force`` does not work.",
        )
        subparser.add_argument(
            "-s", "--stats", dest="stats", action="store_true", help="print statistics for the deleted archive"
        )
        subparser.add_argument(
            "--list", dest="output_list", action="store_true", help="output verbose list of archives it keeps/prunes"
        )
        subparser.add_argument(
            "--keep-within",
            metavar="INTERVAL",
            dest="within",
            type=interval,
            help="keep all archives within this time interval",
        )
        subparser.add_argument(
            "--keep-last",
            "--keep-secondly",
            dest="secondly",
            type=int,
            default=0,
            help="number of secondly archives to keep",
        )
        subparser.add_argument(
            "--keep-minutely", dest="minutely", type=int, default=0, help="number of minutely archives to keep"
        )
        subparser.add_argument(
            "-H", "--keep-hourly", dest="hourly", type=int, default=0, help="number of hourly archives to keep"
        )
        subparser.add_argument(
            "-d", "--keep-daily", dest="daily", type=int, default=0, help="number of daily archives to keep"
        )
        subparser.add_argument(
            "-w", "--keep-weekly", dest="weekly", type=int, default=0, help="number of weekly archives to keep"
        )
        subparser.add_argument(
            "-m", "--keep-monthly", dest="monthly", type=int, default=0, help="number of monthly archives to keep"
        )
        subparser.add_argument(
            "-y", "--keep-yearly", dest="yearly", type=int, default=0, help="number of yearly archives to keep"
        )
        define_archive_filters_group(subparser, sort_by=False, first_last=False)
        subparser.add_argument(
            "-c",
            "--checkpoint-interval",
            metavar="SECONDS",
            dest="checkpoint_interval",
            type=int,
            default=1800,
            help="write checkpoint every SECONDS seconds (Default: 1800)",
        )



from typing import NamedTuple, Tuple, List, Dict, Any, Type, Iterator, BinaryIO

API_VERSION: str

has_seek_hole: bool

class _Chunk(NamedTuple):
    data: bytes
    meta: Dict[str, Any]

def Chunk(data: bytes, **meta) -> Type[_Chunk]: ...
def buzhash(data: bytes, seed: int) -> int: ...
def buzhash_update(sum: int, remove: int, add: int, len: int, seed: int) -> int: ...
def get_chunker(algo: str, *params, **kw) -> Any: ...

fmap_entry = Tuple[int, int, bool]

def sparsemap(fd: BinaryIO = None, fh: int = -1) -> List[fmap_entry]: ...

class ChunkerFailing:
    def __init__(self, block_size: int, map: str) -> None: ...
    def chunkify(self, fd: BinaryIO = None, fh: int = -1) -> Iterator: ...

class ChunkerFixed:
    def __init__(self, block_size: int, header_size: int = 0, sparse: bool = False) -> None: ...
    def chunkify(self, fd: BinaryIO = None, fh: int = -1, fmap: List[fmap_entry] = None) -> Iterator: ...

class Chunker:
    def __init__(
        self, seed: int, chunk_min_exp: int, chunk_max_exp: int, hash_mask_bits: int, hash_window_size: int
    ) -> None: ...
    def chunkify(self, fd: BinaryIO = None, fh: int = -1) -> Iterator: ...



"""
Do benchmarks using pytest-benchmark.

Usage:

    py.test --benchmark-only
"""

import os

import pytest

from .archiver import changedir, cmd
from .item import Item
from ..constants import zeros


@pytest.fixture
def repo_url(request, tmpdir, monkeypatch):
    monkeypatch.setenv("BORG_PASSPHRASE", "123456")
    monkeypatch.setenv("BORG_CHECK_I_KNOW_WHAT_I_AM_DOING", "YES")
    monkeypatch.setenv("BORG_DELETE_I_KNOW_WHAT_I_AM_DOING", "YES")
    monkeypatch.setenv("BORG_UNKNOWN_UNENCRYPTED_REPO_ACCESS_IS_OK", "yes")
    monkeypatch.setenv("BORG_KEYS_DIR", str(tmpdir.join("keys")))
    monkeypatch.setenv("BORG_CACHE_DIR", str(tmpdir.join("cache")))
    yield str(tmpdir.join("repository"))
    tmpdir.remove(rec=1)


@pytest.fixture(params=["none", "repokey-aes-ocb"])
def repo(request, cmd, repo_url):
    cmd(f"--repo={repo_url}", "rcreate", "--encryption", request.param)
    return repo_url


@pytest.fixture(scope="session", params=["zeros", "random"])
def testdata(request, tmpdir_factory):
    count, size = 10, 1000 * 1000
    assert size <= len(zeros)
    p = tmpdir_factory.mktemp("data")
    data_type = request.param
    if data_type == "zeros":
        # do not use a binary zero (\0) to avoid sparse detection
        def data(size):
            return memoryview(zeros)[:size]

    elif data_type == "random":

        def data(size):
            return os.urandom(size)

    else:
        raise ValueError("data_type must be 'random' or 'zeros'.")
    for i in range(count):
        with open(str(p.join(str(i))), "wb") as f:
            f.write(data(size))
    yield str(p)
    p.remove(rec=1)


@pytest.fixture(params=["none", "lz4"])
def repo_archive(request, cmd, repo, testdata):
    archive = "test"
    cmd(f"--repo={repo}", "create", "--compression", request.param, archive, testdata)
    return repo, archive


def test_create_none(benchmark, cmd, repo, testdata):
    result, out = benchmark.pedantic(cmd, (f"--repo={repo}", "create", "--compression", "none", "test", testdata))
    assert result == 0


def test_create_lz4(benchmark, cmd, repo, testdata):
    result, out = benchmark.pedantic(cmd, (f"--repo={repo}", "create", "--compression", "lz4", "test", testdata))
    assert result == 0


def test_extract(benchmark, cmd, repo_archive, tmpdir):
    repo, archive = repo_archive
    with changedir(str(tmpdir)):
        result, out = benchmark.pedantic(cmd, (f"--repo={repo}", "extract", archive))
    assert result == 0


def test_delete(benchmark, cmd, repo_archive):
    repo, archive = repo_archive
    result, out = benchmark.pedantic(cmd, (f"--repo={repo}", "delete", "-a", archive))
    assert result == 0


def test_list(benchmark, cmd, repo_archive):
    repo, archive = repo_archive
    result, out = benchmark(cmd, f"--repo={repo}", "list", archive)
    assert result == 0


def test_info(benchmark, cmd, repo_archive):
    repo, archive = repo_archive
    result, out = benchmark(cmd, f"--repo={repo}", "info", "-a", archive)
    assert result == 0


def test_check(benchmark, cmd, repo_archive):
    repo, archive = repo_archive
    result, out = benchmark(cmd, f"--repo={repo}", "check")
    assert result == 0


def test_help(benchmark, cmd):
    result, out = benchmark(cmd, "help")
    assert result == 0


@pytest.mark.parametrize("type, key, value", [(Item, "size", 1000), (Item, "mode", 0x777), (Item, "deleted", False)])
def test_propdict_attributes(benchmark, type, key, value):
    propdict = type()

    def getattr_setattr(item, key, value):
        setattr(item, key, value)
        assert item.get(key) == value
        assert getattr(item, key) == value
        item.as_dict()

    benchmark(getattr_setattr, propdict, key, value)



import argparse
import textwrap
import sys

from ._common import with_repository, build_matcher
from ..archive import Archive
from ..cache import Cache
from ..constants import *  # NOQA
from ..helpers import ItemFormatter, BaseFormatter, archivename_validator
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


class ListMixIn:
    @with_repository(compatibility=(Manifest.Operation.READ,))
    def do_list(self, args, repository, manifest):
        """List archive contents"""
        matcher = build_matcher(args.patterns, args.paths)
        if args.format is not None:
            format = args.format
        elif args.short:
            format = "{path}{NL}"
        else:
            format = "{mode} {user:6} {group:6} {size:8} {mtime} {path}{extra}{NL}"

        def _list_inner(cache):
            archive = Archive(manifest, args.name, cache=cache)

            formatter = ItemFormatter(archive, format, json_lines=args.json_lines)
            for item in archive.iter_items(lambda item: matcher.match(item.path)):
                sys.stdout.write(formatter.format_item(item))

        # Only load the cache if it will be used
        if ItemFormatter.format_needs_cache(format):
            with Cache(repository, manifest, lock_wait=self.lock_wait) as cache:
                _list_inner(cache)
        else:
            _list_inner(cache=None)

        return self.exit_code

    def build_parser_list(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog, define_exclusion_group

        list_epilog = (
            process_epilog(
                """
        This command lists the contents of an archive.

        For more help on include/exclude patterns, see the :ref:`borg_patterns` command output.

        .. man NOTES

        The FORMAT specifier syntax
        +++++++++++++++++++++++++++

        The ``--format`` option uses python's `format string syntax
        <https://docs.python.org/3.9/library/string.html#formatstrings>`_.

        Examples:
        ::

            $ borg list --format '{mode} {user:6} {group:6} {size:8} {mtime} {path}{extra}{NL}' ArchiveFoo
            -rw-rw-r-- user   user       1024 Thu, 2021-12-09 10:22:17 file-foo
            ...

            # {VAR:<NUMBER} - pad to NUMBER columns left-aligned.
            # {VAR:>NUMBER} - pad to NUMBER columns right-aligned.
            $ borg list --format '{mode} {user:>6} {group:>6} {size:<8} {mtime} {path}{extra}{NL}' ArchiveFoo
            -rw-rw-r--   user   user 1024     Thu, 2021-12-09 10:22:17 file-foo
            ...

        The following keys are always available:


        """
            )
            + BaseFormatter.keys_help()
            + textwrap.dedent(
                """

        Keys available only when listing files in an archive:

        """
            )
            + ItemFormatter.keys_help()
        )
        subparser = subparsers.add_parser(
            "list",
            parents=[common_parser],
            add_help=False,
            description=self.do_list.__doc__,
            epilog=list_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="list archive contents",
        )
        subparser.set_defaults(func=self.do_list)
        subparser.add_argument(
            "--short", dest="short", action="store_true", help="only print file/directory names, nothing else"
        )
        subparser.add_argument(
            "--format",
            metavar="FORMAT",
            dest="format",
            help="specify format for file listing "
            '(default: "{mode} {user:6} {group:6} {size:8} {mtime} {path}{extra}{NL}")',
        )
        subparser.add_argument(
            "--json-lines",
            action="store_true",
            help="Format output as JSON Lines. "
            "The form of ``--format`` is ignored, "
            "but keys used in it are added to the JSON output. "
            "Some keys are always present. Note: JSON can only represent text.",
        )
        subparser.add_argument("name", metavar="NAME", type=archivename_validator, help="specify the archive name")
        subparser.add_argument(
            "paths", metavar="PATH", nargs="*", type=str, help="paths to list; patterns are supported"
        )
        define_exclusion_group(subparser)



"""
wrapping msgpack
================

We wrap msgpack here the way we need it - to avoid having lots of clutter in the calling code.

Packing
-------
- use_bin_type = True (used by borg since borg 2.0)
  This is used to generate output according to new msgpack 2.0 spec.
  This cleanly keeps bytes and str types apart.

- use_bin_type = False (used by borg < 1.3)
  This creates output according to the older msgpack spec.
  BAD: str and bytes were packed into same "raw" representation.

- unicode_errors = 'surrogateescape'
  Guess backup applications are one of the rare cases when this needs to be used.
  It is needed because borg also needs to deal with data that does not cleanly encode/decode using utf-8.
  There's a lot of crap out there, e.g. in filenames and as a backup tool, we must keep them as good as possible.

Unpacking
---------
- raw = False (used by borg since borg 2.0)
  We already can use this with borg 2.0 due to the type conversion to the desired type in item.py update_internal
  methods. This type conversion code can be removed in future, when we do not have to deal with data any more
  that was packed the old way.
  It will then unpack according to the msgpack 2.0 spec format and directly output bytes or str.

- raw = True (the old way, used by borg < 1.3)

- unicode_errors = 'surrogateescape' -> see description above (will be used when raw is False).

As of borg 2.0, we have fixed most of the msgpack str/bytes mess, #968.
Borg now still needs to **read** old repos, archives, keys, ... so we can not yet fix it completely.
But from now on, borg only **writes** new data according to the new msgpack 2.0 spec,
thus we can remove some legacy support in a later borg release (some places are marked with "legacy").

current way in msgpack terms
----------------------------

- pack with use_bin_type=True (according to msgpack 2.0 spec)
- packs str -> raw and bytes -> bin
- unpack with raw=False (according to msgpack 2.0 spec, using unicode_errors='surrogateescape')
- unpacks bin to bytes and raw to str (thus we need to convert to desired type if we want bytes from "raw")
"""

from .datastruct import StableDict
from ..constants import *  # NOQA

from msgpack import Packer as mp_Packer
from msgpack import packb as mp_packb
from msgpack import pack as mp_pack
from msgpack import Unpacker as mp_Unpacker
from msgpack import unpackb as mp_unpackb
from msgpack import unpack as mp_unpack
from msgpack import version as mp_version

from msgpack import ExtType, Timestamp
from msgpack import OutOfData


version = mp_version

USE_BIN_TYPE = True
RAW = False
UNICODE_ERRORS = "surrogateescape"


class PackException(Exception):
    """Exception while msgpack packing"""


class UnpackException(Exception):
    """Exception while msgpack unpacking"""


class Packer(mp_Packer):
    def __init__(
        self,
        *,
        default=None,
        unicode_errors=UNICODE_ERRORS,
        use_single_float=False,
        autoreset=True,
        use_bin_type=USE_BIN_TYPE,
        strict_types=False
    ):
        assert unicode_errors == UNICODE_ERRORS
        super().__init__(
            default=default,
            unicode_errors=unicode_errors,
            use_single_float=use_single_float,
            autoreset=autoreset,
            use_bin_type=use_bin_type,
            strict_types=strict_types,
        )

    def pack(self, obj):
        try:
            return super().pack(obj)
        except Exception as e:
            raise PackException(e)


def packb(o, *, use_bin_type=USE_BIN_TYPE, unicode_errors=UNICODE_ERRORS, **kwargs):
    assert unicode_errors == UNICODE_ERRORS
    try:
        return mp_packb(o, use_bin_type=use_bin_type, unicode_errors=unicode_errors, **kwargs)
    except Exception as e:
        raise PackException(e)


def pack(o, stream, *, use_bin_type=USE_BIN_TYPE, unicode_errors=UNICODE_ERRORS, **kwargs):
    assert unicode_errors == UNICODE_ERRORS
    try:
        return mp_pack(o, stream, use_bin_type=use_bin_type, unicode_errors=unicode_errors, **kwargs)
    except Exception as e:
        raise PackException(e)


class Unpacker(mp_Unpacker):
    def __init__(
        self,
        file_like=None,
        *,
        read_size=0,
        use_list=True,
        raw=RAW,
        object_hook=None,
        object_pairs_hook=None,
        list_hook=None,
        unicode_errors=UNICODE_ERRORS,
        max_buffer_size=0,
        ext_hook=ExtType,
        strict_map_key=False
    ):
        assert raw == RAW
        assert unicode_errors == UNICODE_ERRORS
        kw = dict(
            file_like=file_like,
            read_size=read_size,
            use_list=use_list,
            raw=raw,
            object_hook=object_hook,
            object_pairs_hook=object_pairs_hook,
            list_hook=list_hook,
            unicode_errors=unicode_errors,
            max_buffer_size=max_buffer_size,
            ext_hook=ext_hook,
            strict_map_key=strict_map_key,
        )
        super().__init__(**kw)

    def unpack(self):
        try:
            return super().unpack()
        except OutOfData:
            raise
        except Exception as e:
            raise UnpackException(e)

    def __next__(self):
        try:
            return super().__next__()
        except StopIteration:
            raise
        except Exception as e:
            raise UnpackException(e)

    next = __next__


def unpackb(packed, *, raw=RAW, unicode_errors=UNICODE_ERRORS, strict_map_key=False, **kwargs):
    assert raw == RAW
    assert unicode_errors == UNICODE_ERRORS
    try:
        kw = dict(raw=raw, unicode_errors=unicode_errors, strict_map_key=strict_map_key)
        kw.update(kwargs)
        return mp_unpackb(packed, **kw)
    except Exception as e:
        raise UnpackException(e)


def unpack(stream, *, raw=RAW, unicode_errors=UNICODE_ERRORS, strict_map_key=False, **kwargs):
    assert raw == RAW
    assert unicode_errors == UNICODE_ERRORS
    try:
        kw = dict(raw=raw, unicode_errors=unicode_errors, strict_map_key=strict_map_key)
        kw.update(kwargs)
        return mp_unpack(stream, **kw)
    except Exception as e:
        raise UnpackException(e)


# msgpacking related utilities -----------------------------------------------


def is_slow_msgpack():
    import msgpack
    import msgpack.fallback

    return msgpack.Packer is msgpack.fallback.Packer


def is_supported_msgpack():
    # DO NOT CHANGE OR REMOVE! See also requirements and comments in setup.py.
    import msgpack

    return (1, 0, 3) <= msgpack.version <= (
        1,
        0,
        4,
    ) and msgpack.version not in []  # < add bad releases here to deny list


def get_limited_unpacker(kind):
    """return a limited Unpacker because we should not trust msgpack data received from remote"""
    # Note: msgpack >= 0.6.1 auto-computes DoS-safe max values from len(data) for
    #       unpack(data) or from max_buffer_size for Unpacker(max_buffer_size=N).
    args = dict(use_list=False, max_buffer_size=3 * max(BUFSIZE, MAX_OBJECT_SIZE))  # return tuples, not lists
    if kind in ("server", "client"):
        pass  # nothing special
    elif kind in ("manifest", "key"):
        args.update(dict(use_list=True, object_hook=StableDict))  # default value
    else:
        raise ValueError('kind must be "server", "client", "manifest" or "key"')
    return Unpacker(**args)


def int_to_timestamp(ns):
    assert isinstance(ns, int)
    return Timestamp.from_unix_nano(ns)


def timestamp_to_int(ts):
    assert isinstance(ts, Timestamp)
    return ts.to_unix_nano()



from typing import Any, Type, Dict, Tuple

API_VERSION: str

def get_compressor(name: str, **kwargs) -> Any: ...

class CompressionSpec:
    def __init__(self, spec: str) -> None: ...
    @property
    def compressor(self) -> Any: ...
    inner: CompressionSpec

class Compressor:
    def __init__(self, name: Any = ..., **kwargs) -> None: ...
    def compress(self, meta: Dict, data: bytes) -> Tuple[Dict, bytes]: ...
    def decompress(self, meta: Dict, data: bytes) -> Tuple[Dict, bytes]: ...
    @staticmethod
    def detect(data: bytes) -> Any: ...

class CompressorBase:
    ID: bytes = ...
    name: str = ...
    @classmethod
    def detect(self, data: bytes) -> bool: ...
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    def decide(self, data: bytes) -> Any: ...
    def compress(self, data: bytes) -> bytes: ...
    def decompress(self, data: bytes) -> bytes: ...

class Auto(CompressorBase):
    def __init__(self, compressor: Any) -> None: ...

class DecidingCompressor(CompressorBase):
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    def decide_compress(self, data: bytes) -> Any: ...

class CNONE(CompressorBase):
    def __init__(self, level: int = ..., **kwargs) -> None: ...

class ObfuscateSize(CompressorBase):
    def __init__(self, level: int = ..., compressor: Any = ...) -> None: ...

class ZLIB_legacy(CompressorBase):
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    level: int

class ZLIB(CompressorBase):
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    level: int

class LZ4(DecidingCompressor):
    def __init__(self, level: int = ..., **kwargs) -> None: ...

class LZMA(DecidingCompressor):
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    level: int

class ZSTD(DecidingCompressor):
    def __init__(self, level: int = ..., **kwargs) -> None: ...
    level: int

LZ4_COMPRESSOR: Type[LZ4]
NONE_COMPRESSOR: Type[CNONE]

COMPRESSOR_TABLE: dict



import argparse
from ._common import with_repository
from ..archive import ArchiveChecker
from ..constants import *  # NOQA
from ..helpers import EXIT_SUCCESS, EXIT_WARNING, EXIT_ERROR
from ..helpers import yes

from ..logger import create_logger

logger = create_logger()


class CheckMixIn:
    @with_repository(exclusive=True, manifest=False)
    def do_check(self, args, repository):
        """Check repository consistency"""
        if args.repair:
            msg = (
                "This is a potentially dangerous function.\n"
                "check --repair might lead to data loss (for kinds of corruption it is not\n"
                "capable of dealing with). BE VERY CAREFUL!\n"
                "\n"
                "Type 'YES' if you understand this and want to continue: "
            )
            if not yes(
                msg,
                false_msg="Aborting.",
                invalid_msg="Invalid answer, aborting.",
                truish=("YES",),
                retry=False,
                env_var_override="BORG_CHECK_I_KNOW_WHAT_I_AM_DOING",
            ):
                return EXIT_ERROR
        if args.repo_only and any((args.verify_data, args.first, args.last, args.match_archives)):
            self.print_error(
                "--repository-only contradicts --first, --last, -a / --match-archives and --verify-data arguments."
            )
            return EXIT_ERROR
        if args.repair and args.max_duration:
            self.print_error("--repair does not allow --max-duration argument.")
            return EXIT_ERROR
        if args.max_duration and not args.repo_only:
            # when doing a partial repo check, we can only check crc32 checksums in segment files,
            # we can't build a fresh repo index in memory to verify the on-disk index against it.
            # thus, we should not do an archives check based on a unknown-quality on-disk repo index.
            # also, there is no max_duration support in the archives check code anyway.
            self.print_error("--repository-only is required for --max-duration support.")
            return EXIT_ERROR
        if not args.archives_only:
            if not repository.check(repair=args.repair, max_duration=args.max_duration):
                return EXIT_WARNING
        if not args.repo_only and not ArchiveChecker().check(
            repository,
            verify_data=args.verify_data,
            repair=args.repair,
            match=args.match_archives,
            sort_by=args.sort_by or "ts",
            first=args.first,
            last=args.last,
            older=args.older,
            newer=args.newer,
            oldest=args.oldest,
            newest=args.newest,
        ):
            return EXIT_WARNING
        return EXIT_SUCCESS

    def build_parser_check(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog
        from ._common import define_archive_filters_group

        check_epilog = process_epilog(
            """
        The check command verifies the consistency of a repository and the corresponding archives.

        check --repair is a potentially dangerous function and might lead to data loss
        (for kinds of corruption it is not capable of dealing with). BE VERY CAREFUL!

        Pursuant to the previous warning it is also highly recommended to test the
        reliability of the hardware running this software with stress testing software
        such as memory testers. Unreliable hardware can also lead to data loss especially
        when this command is run in repair mode.

        First, the underlying repository data files are checked:

        - For all segments, the segment magic header is checked.
        - For all objects stored in the segments, all metadata (e.g. CRC and size) and
          all data is read. The read data is checked by size and CRC. Bit rot and other
          types of accidental damage can be detected this way.
        - In repair mode, if an integrity error is detected in a segment, try to recover
          as many objects from the segment as possible.
        - In repair mode, make sure that the index is consistent with the data stored in
          the segments.
        - If checking a remote repo via ``ssh:``, the repo check is executed on the server
          without causing significant network traffic.
        - The repository check can be skipped using the ``--archives-only`` option.
        - A repository check can be time consuming. Partial checks are possible with the
          ``--max-duration`` option.

        Second, the consistency and correctness of the archive metadata is verified:

        - Is the repo manifest present? If not, it is rebuilt from archive metadata
          chunks (this requires reading and decrypting of all metadata and data).
        - Check if archive metadata chunk is present; if not, remove archive from manifest.
        - For all files (items) in the archive, for all chunks referenced by these
          files, check if chunk is present. In repair mode, if a chunk is not present,
          replace it with a same-size replacement chunk of zeroes. If a previously lost
          chunk reappears (e.g. via a later backup), in repair mode the all-zero replacement
          chunk will be replaced by the correct chunk. This requires reading of archive and
          file metadata, but not data.
        - In repair mode, when all the archives were checked, orphaned chunks are deleted
          from the repo. One cause of orphaned chunks are input file related errors (like
          read errors) in the archive creation process.
        - In verify-data mode, a complete cryptographic verification of the archive data
          integrity is performed. This conflicts with ``--repository-only`` as this mode
          only makes sense if the archive checks are enabled. The full details of this mode
          are documented below.
        - If checking a remote repo via ``ssh:``, the archive check is executed on the
          client machine because it requires decryption, and this is always done client-side
          as key access is needed.
        - The archive checks can be time consuming; they can be skipped using the
          ``--repository-only`` option.

        The ``--max-duration`` option can be used to split a long-running repository check
        into multiple partial checks. After the given number of seconds the check is
        interrupted. The next partial check will continue where the previous one stopped,
        until the complete repository has been checked. Example: Assuming a complete check took 7
        hours, then running a daily check with --max-duration=3600 (1 hour) resulted in one
        completed check per week.

        Attention: A partial --repository-only check can only do way less checking than a full
        --repository-only check: only the non-cryptographic checksum checks on segment file
        entries are done, while a full --repository-only check would also do a repo index check.
        A partial check cannot be combined with the ``--repair`` option. Partial checks
        may therefore be useful only with very large repositories where a full check would take
        too long.
        Doing a full repository check aborts a partial check; the next partial check will restart
        from the beginning.

        The ``--verify-data`` option will perform a full integrity verification (as opposed to
        checking the CRC32 of the segment) of data, which means reading the data from the
        repository, decrypting and decompressing it. This is a cryptographic verification,
        which will detect (accidental) corruption. For encrypted repositories it is
        tamper-resistant as well, unless the attacker has access to the keys. It is also very
        slow.
        """
        )
        subparser = subparsers.add_parser(
            "check",
            parents=[common_parser],
            add_help=False,
            description=self.do_check.__doc__,
            epilog=check_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="verify repository",
        )
        subparser.set_defaults(func=self.do_check)
        subparser.add_argument(
            "--repository-only", dest="repo_only", action="store_true", help="only perform repository checks"
        )
        subparser.add_argument(
            "--archives-only", dest="archives_only", action="store_true", help="only perform archives checks"
        )
        subparser.add_argument(
            "--verify-data",
            dest="verify_data",
            action="store_true",
            help="perform cryptographic archive data integrity verification " "(conflicts with ``--repository-only``)",
        )
        subparser.add_argument(
            "--repair", dest="repair", action="store_true", help="attempt to repair any inconsistencies found"
        )
        subparser.add_argument(
            "--max-duration",
            metavar="SECONDS",
            dest="max_duration",
            type=int,
            default=0,
            help="do only a partial repo check for max. SECONDS seconds (Default: unlimited)",
        )
        define_archive_filters_group(subparser)



import pytest

from ..helpers.datastruct import EfficientCollectionQueue


class TestEfficientQueue:
    def test_base_usage(self):
        queue = EfficientCollectionQueue(100, bytes)
        assert queue.peek_front() == b""
        queue.push_back(b"1234")
        assert queue.peek_front() == b"1234"
        assert len(queue) == 4
        assert queue
        queue.pop_front(4)
        assert queue.peek_front() == b""
        assert len(queue) == 0
        assert not queue

    def test_usage_with_arrays(self):
        queue = EfficientCollectionQueue(100, list)
        assert queue.peek_front() == []
        queue.push_back([1, 2, 3, 4])
        assert queue.peek_front() == [1, 2, 3, 4]
        assert len(queue) == 4
        assert queue
        queue.pop_front(4)
        assert queue.peek_front() == []
        assert len(queue) == 0
        assert not queue

    def test_chunking(self):
        queue = EfficientCollectionQueue(2, bytes)
        queue.push_back(b"1")
        queue.push_back(b"23")
        queue.push_back(b"4567")
        assert len(queue) == 7
        assert queue.peek_front() == b"12"
        queue.pop_front(3)
        assert queue.peek_front() == b"4"
        queue.pop_front(1)
        assert queue.peek_front() == b"56"
        queue.pop_front(2)
        assert len(queue) == 1
        assert queue
        with pytest.raises(EfficientCollectionQueue.SizeUnderflow):
            queue.pop_front(2)
        assert queue.peek_front() == b"7"
        queue.pop_front(1)
        assert queue.peek_front() == b""
        assert len(queue) == 0
        assert not queue



import re

import pytest

from ..helpers import shellpattern


def check(path, pattern):
    compiled = re.compile(shellpattern.translate(pattern))

    return bool(compiled.match(path))


@pytest.mark.parametrize(
    "path, patterns",
    [
        # Literal string
        ("foo/bar", ["foo/bar"]),
        ("foo\\bar", ["foo\\bar"]),
        # Non-ASCII
        ("foo/c/\u0152/e/bar", ["foo/*/\u0152/*/bar", "*/*/\u0152/*/*", "**/\u0152/*/*"]),
        ("\u00e4\u00f6\u00dc", ["???", "*", "\u00e4\u00f6\u00dc", "[\u00e4][\u00f6][\u00dc]"]),
        # Question mark
        ("foo", ["fo?"]),
        ("foo", ["f?o"]),
        ("foo", ["f??"]),
        ("foo", ["?oo"]),
        ("foo", ["?o?"]),
        ("foo", ["??o"]),
        ("foo", ["???"]),
        # Single asterisk
        ("", ["*"]),
        ("foo", ["*", "**", "***"]),
        ("foo", ["foo*"]),
        ("foobar", ["foo*"]),
        ("foobar", ["foo*bar"]),
        ("foobarbaz", ["foo*baz"]),
        ("bar", ["*bar"]),
        ("foobar", ["*bar"]),
        ("foo/bar", ["foo/*bar"]),
        ("foo/bar", ["foo/*ar"]),
        ("foo/bar", ["foo/*r"]),
        ("foo/bar", ["foo/*"]),
        ("foo/bar", ["foo*/bar"]),
        ("foo/bar", ["fo*/bar"]),
        ("foo/bar", ["f*/bar"]),
        ("foo/bar", ["*/bar"]),
        # Double asterisk (matches 0..n directory layers)
        ("foo/bar", ["foo/**/bar"]),
        ("foo/1/bar", ["foo/**/bar"]),
        ("foo/1/22/333/bar", ["foo/**/bar"]),
        ("foo/", ["foo/**/"]),
        ("foo/1/", ["foo/**/"]),
        ("foo/1/22/333/", ["foo/**/"]),
        ("bar", ["**/bar"]),
        ("1/bar", ["**/bar"]),
        ("1/22/333/bar", ["**/bar"]),
        ("foo/bar/baz", ["foo/**/*"]),
        # Set
        ("foo1", ["foo[12]"]),
        ("foo2", ["foo[12]"]),
        ("foo2/bar", ["foo[12]/*"]),
        ("f??f", ["f??f", "f[?][?]f"]),
        ("foo]", ["foo[]]"]),
        # Inverted set
        ("foo3", ["foo[!12]"]),
        ("foo^", ["foo[^!]"]),
        ("foo!", ["foo[^!]"]),
    ],
)
def test_match(path, patterns):
    for p in patterns:
        assert check(path, p)


@pytest.mark.parametrize(
    "path, patterns",
    [
        ("", ["?", "[]"]),
        ("foo", ["foo?"]),
        ("foo", ["?foo"]),
        ("foo", ["f?oo"]),
        # do not match path separator
        ("foo/ar", ["foo?ar"]),
        # do not match/cross over os.path.sep
        ("foo/bar", ["*"]),
        ("foo/bar", ["foo*bar"]),
        ("foo/bar", ["foo*ar"]),
        ("foo/bar", ["fo*bar"]),
        ("foo/bar", ["fo*ar"]),
        # Double asterisk
        ("foobar", ["foo/**/bar"]),
        # Two asterisks without slash do not match directory separator
        ("foo/bar", ["**"]),
        # Double asterisk not matching filename
        ("foo/bar", ["**/"]),
        # Set
        ("foo3", ["foo[12]"]),
        # Inverted set
        ("foo1", ["foo[!12]"]),
        ("foo2", ["foo[!12]"]),
    ],
)
def test_mismatch(path, patterns):
    for p in patterns:
        assert not check(path, p)


def test_match_end():
    regex = shellpattern.translate("*-home")  # default is match_end == string end
    assert re.match(regex, "2017-07-03-home")
    assert not re.match(regex, "2017-07-03-home.checkpoint")

    match_end = r"(%s)?\Z" % r"\.checkpoint(\.\d+)?"  # with/without checkpoint ending
    regex = shellpattern.translate("*-home", match_end=match_end)
    assert re.match(regex, "2017-07-03-home")
    assert re.match(regex, "2017-07-03-home.checkpoint")



import argparse
import functools
import os

from ..constants import *  # NOQA
from ..crypto.key import AESOCBRepoKey, CHPORepoKey, Blake2AESOCBRepoKey, Blake2CHPORepoKey
from ..crypto.key import AESOCBKeyfileKey, CHPOKeyfileKey, Blake2AESOCBKeyfileKey, Blake2CHPOKeyfileKey
from ..crypto.keymanager import KeyManager
from ..manifest import Manifest

from ._common import with_repository

from ..logger import create_logger

logger = create_logger(__name__)


class KeysMixIn:
    @with_repository(compatibility=(Manifest.Operation.CHECK,))
    def do_change_passphrase(self, args, repository, manifest):
        """Change repository key file passphrase"""
        key = manifest.key
        if not hasattr(key, "change_passphrase"):
            print("This repository is not encrypted, cannot change the passphrase.")
            return EXIT_ERROR
        key.change_passphrase()
        logger.info("Key updated")
        if hasattr(key, "find_key"):
            # print key location to make backing it up easier
            logger.info("Key location: %s", key.find_key())
        return EXIT_SUCCESS

    @with_repository(exclusive=True, manifest=True, cache=True, compatibility=(Manifest.Operation.CHECK,))
    def do_change_location(self, args, repository, manifest, cache):
        """Change repository key location"""
        key = manifest.key
        if not hasattr(key, "change_passphrase"):
            print("This repository is not encrypted, cannot change the key location.")
            return EXIT_ERROR

        if args.key_mode == "keyfile":
            if isinstance(key, AESOCBRepoKey):
                key_new = AESOCBKeyfileKey(repository)
            elif isinstance(key, CHPORepoKey):
                key_new = CHPOKeyfileKey(repository)
            elif isinstance(key, Blake2AESOCBRepoKey):
                key_new = Blake2AESOCBKeyfileKey(repository)
            elif isinstance(key, Blake2CHPORepoKey):
                key_new = Blake2CHPOKeyfileKey(repository)
            else:
                print("Change not needed or not supported.")
                return EXIT_WARNING
        if args.key_mode == "repokey":
            if isinstance(key, AESOCBKeyfileKey):
                key_new = AESOCBRepoKey(repository)
            elif isinstance(key, CHPOKeyfileKey):
                key_new = CHPORepoKey(repository)
            elif isinstance(key, Blake2AESOCBKeyfileKey):
                key_new = Blake2AESOCBRepoKey(repository)
            elif isinstance(key, Blake2CHPOKeyfileKey):
                key_new = Blake2CHPORepoKey(repository)
            else:
                print("Change not needed or not supported.")
                return EXIT_WARNING

        for name in ("repository_id", "crypt_key", "id_key", "chunk_seed", "tam_required", "sessionid", "cipher"):
            value = getattr(key, name)
            setattr(key_new, name, value)

        key_new.target = key_new.get_new_target(args)
        # save with same passphrase and algorithm
        key_new.save(key_new.target, key._passphrase, create=True, algorithm=key._encrypted_key_algorithm)

        # rewrite the manifest with the new key, so that the key-type byte of the manifest changes
        manifest.key = key_new
        manifest.repo_objs.key = key_new
        manifest.write()
        repository.commit(compact=False)

        # we need to rewrite cache config and security key-type info,
        # so that the cached key-type will match the repo key-type.
        cache.begin_txn()  # need to start a cache transaction, otherwise commit() does nothing.
        cache.key = key_new
        cache.commit()

        loc = key_new.find_key() if hasattr(key_new, "find_key") else None
        if args.keep:
            logger.info(f"Key copied to {loc}")
        else:
            key.remove(key.target)  # remove key from current location
            logger.info(f"Key moved to {loc}")

        return EXIT_SUCCESS

    @with_repository(lock=False, exclusive=False, manifest=False, cache=False)
    def do_key_export(self, args, repository):
        """Export the repository key for backup"""
        manager = KeyManager(repository)
        manager.load_keyblob()
        if args.paper:
            manager.export_paperkey(args.path)
        else:
            try:
                if os.path.isdir(args.path):
                    raise IsADirectoryError
                if args.qr:
                    manager.export_qr(args.path)
                else:
                    manager.export(args.path)
            except IsADirectoryError:
                self.print_error(f"'{args.path}' must be a file, not a directory")
                return EXIT_ERROR
        return EXIT_SUCCESS

    @with_repository(lock=False, exclusive=False, manifest=False, cache=False)
    def do_key_import(self, args, repository):
        """Import the repository key from backup"""
        manager = KeyManager(repository)
        if args.paper:
            if args.path:
                self.print_error("with --paper import from file is not supported")
                return EXIT_ERROR
            manager.import_paperkey(args)
        else:
            if not args.path:
                self.print_error("input file to import key from expected")
                return EXIT_ERROR
            if args.path != "-" and not os.path.exists(args.path):
                self.print_error("input file does not exist: " + args.path)
                return EXIT_ERROR
            manager.import_keyfile(args)
        return EXIT_SUCCESS

    def build_parser_keys(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        subparser = subparsers.add_parser(
            "key",
            parents=[mid_common_parser],
            add_help=False,
            description="Manage a keyfile or repokey of a repository",
            epilog="",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="manage repository key",
        )

        key_parsers = subparser.add_subparsers(title="required arguments", metavar="<command>")
        subparser.set_defaults(fallback_func=functools.partial(self.do_subcommand_help, subparser))

        key_export_epilog = process_epilog(
            """
        If repository encryption is used, the repository is inaccessible
        without the key. This command allows one to back up this essential key.
        Note that the backup produced does not include the passphrase itself
        (i.e. the exported key stays encrypted). In order to regain access to a
        repository, one needs both the exported key and the original passphrase.

        There are three backup formats. The normal backup format is suitable for
        digital storage as a file. The ``--paper`` backup format is optimized
        for printing and typing in while importing, with per line checks to
        reduce problems with manual input. The ``--qr-html`` creates a printable
        HTML template with a QR code and a copy of the ``--paper``-formatted key.

        For repositories using keyfile encryption the key is saved locally
        on the system that is capable of doing backups. To guard against loss
        of this key, the key needs to be backed up independently of the main
        data backup.

        For repositories using the repokey encryption the key is saved in the
        repository in the config file. A backup is thus not strictly needed,
        but guards against the repository becoming inaccessible if the file
        is damaged for some reason.

        Examples::

            borg key export /path/to/repo > encrypted-key-backup
            borg key export --paper /path/to/repo > encrypted-key-backup.txt
            borg key export --qr-html /path/to/repo > encrypted-key-backup.html
            # Or pass the output file as an argument instead of redirecting stdout:
            borg key export /path/to/repo encrypted-key-backup
            borg key export --paper /path/to/repo encrypted-key-backup.txt
            borg key export --qr-html /path/to/repo encrypted-key-backup.html


        """
        )
        subparser = key_parsers.add_parser(
            "export",
            parents=[common_parser],
            add_help=False,
            description=self.do_key_export.__doc__,
            epilog=key_export_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="export repository key for backup",
        )
        subparser.set_defaults(func=self.do_key_export)
        subparser.add_argument("path", metavar="PATH", nargs="?", type=str, help="where to store the backup")
        subparser.add_argument(
            "--paper",
            dest="paper",
            action="store_true",
            help="Create an export suitable for printing and later type-in",
        )
        subparser.add_argument(
            "--qr-html",
            dest="qr",
            action="store_true",
            help="Create an html file suitable for printing and later type-in or qr scan",
        )

        key_import_epilog = process_epilog(
            """
        This command restores a key previously backed up with the export command.

        If the ``--paper`` option is given, the import will be an interactive
        process in which each line is checked for plausibility before
        proceeding to the next line. For this format PATH must not be given.

        For repositories using keyfile encryption, the key file which ``borg key
        import`` writes to depends on several factors. If the ``BORG_KEY_FILE``
        environment variable is set and non-empty, ``borg key import`` creates
        or overwrites that file named by ``$BORG_KEY_FILE``. Otherwise, ``borg
        key import`` searches in the ``$BORG_KEYS_DIR`` directory for a key file
        associated with the repository. If a key file is found in
        ``$BORG_KEYS_DIR``, ``borg key import`` overwrites it; otherwise, ``borg
        key import`` creates a new key file in ``$BORG_KEYS_DIR``.
        """
        )
        subparser = key_parsers.add_parser(
            "import",
            parents=[common_parser],
            add_help=False,
            description=self.do_key_import.__doc__,
            epilog=key_import_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="import repository key from backup",
        )
        subparser.set_defaults(func=self.do_key_import)
        subparser.add_argument(
            "path", metavar="PATH", nargs="?", type=str, help="path to the backup ('-' to read from stdin)"
        )
        subparser.add_argument(
            "--paper",
            dest="paper",
            action="store_true",
            help="interactively import from a backup done with ``--paper``",
        )

        change_passphrase_epilog = process_epilog(
            """
        The key files used for repository encryption are optionally passphrase
        protected. This command can be used to change this passphrase.

        Please note that this command only changes the passphrase, but not any
        secret protected by it (like e.g. encryption/MAC keys or chunker seed).
        Thus, changing the passphrase after passphrase and borg key got compromised
        does not protect future (nor past) backups to the same repository.
        """
        )
        subparser = key_parsers.add_parser(
            "change-passphrase",
            parents=[common_parser],
            add_help=False,
            description=self.do_change_passphrase.__doc__,
            epilog=change_passphrase_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="change repository passphrase",
        )
        subparser.set_defaults(func=self.do_change_passphrase)

        change_location_epilog = process_epilog(
            """
        Change the location of a borg key. The key can be stored at different locations:

        - keyfile: locally, usually in the home directory
        - repokey: inside the repo (in the repo config)

        Please note:

        This command does NOT change the crypto algorithms, just the key location,
        thus you must ONLY give the key location (keyfile or repokey).
        """
        )
        subparser = key_parsers.add_parser(
            "change-location",
            parents=[common_parser],
            add_help=False,
            description=self.do_change_location.__doc__,
            epilog=change_location_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="change key location",
        )
        subparser.set_defaults(func=self.do_change_location)
        subparser.add_argument(
            "key_mode", metavar="KEY_LOCATION", choices=("repokey", "keyfile"), help="select key location"
        )
        subparser.add_argument(
            "--keep",
            dest="keep",
            action="store_true",
            help="keep the key also at the current location (default: remove it)",
        )



import argparse
import functools
import os
import textwrap

import borg
from ..archive import Archive
from ..constants import *  # NOQA
from ..cache import Cache, assert_secure
from ..helpers import Error
from ..helpers import SortBySpec, positive_int_validator, location_validator, Location, relative_time_marker_validator
from ..helpers.nanorst import rst_to_terminal
from ..manifest import Manifest, AI_HUMAN_SORT_KEYS
from ..patterns import PatternMatcher
from ..remote import RemoteRepository
from ..repository import Repository
from ..repoobj import RepoObj, RepoObj1
from ..patterns import (
    ArgparsePatternAction,
    ArgparseExcludeFileAction,
    ArgparsePatternFileAction,
    parse_exclude_pattern,
)


from ..logger import create_logger

logger = create_logger(__name__)


def get_repository(location, *, create, exclusive, lock_wait, lock, append_only, make_parent_dirs, storage_quota, args):
    if location.proto == "ssh":
        repository = RemoteRepository(
            location,
            create=create,
            exclusive=exclusive,
            lock_wait=lock_wait,
            lock=lock,
            append_only=append_only,
            make_parent_dirs=make_parent_dirs,
            args=args,
        )

    else:
        repository = Repository(
            location.path,
            create=create,
            exclusive=exclusive,
            lock_wait=lock_wait,
            lock=lock,
            append_only=append_only,
            make_parent_dirs=make_parent_dirs,
            storage_quota=storage_quota,
        )
    return repository


def compat_check(*, create, manifest, key, cache, compatibility, decorator_name):
    if not create and (manifest or key or cache):
        if compatibility is None:
            raise AssertionError(f"{decorator_name} decorator used without compatibility argument")
        if type(compatibility) is not tuple:
            raise AssertionError(f"{decorator_name} decorator compatibility argument must be of type tuple")
    else:
        if compatibility is not None:
            raise AssertionError(
                f"{decorator_name} called with compatibility argument, " f"but would not check {compatibility!r}"
            )
        if create:
            compatibility = Manifest.NO_OPERATION_CHECK
    return compatibility


def with_repository(
    create=False, lock=True, exclusive=False, manifest=True, cache=False, secure=True, compatibility=None
):
    """
    Method decorator for subcommand-handling methods: do_XYZ(self, args, repository, …)

    If a parameter (where allowed) is a str the attribute named of args is used instead.
    :param create: create repository
    :param lock: lock repository
    :param exclusive: (bool) lock repository exclusively (for writing)
    :param manifest: load manifest and repo_objs (key), pass them as keyword arguments
    :param cache: open cache, pass it as keyword argument (implies manifest)
    :param secure: do assert_secure after loading manifest
    :param compatibility: mandatory if not create and (manifest or cache), specifies mandatory feature categories to check
    """
    # Note: with_repository decorator does not have a "key" argument (yet?)
    compatibility = compat_check(
        create=create,
        manifest=manifest,
        key=manifest,
        cache=cache,
        compatibility=compatibility,
        decorator_name="with_repository",
    )

    # To process the `--bypass-lock` option if specified, we need to
    # modify `lock` inside `wrapper`. Therefore we cannot use the
    # `nonlocal` statement to access `lock` as modifications would also
    # affect the scope outside of `wrapper`. Subsequent calls would
    # only see the overwritten value of `lock`, not the original one.
    # The solution is to define a place holder variable `_lock` to
    # propagate the value into `wrapper`.
    _lock = lock

    def decorator(method):
        @functools.wraps(method)
        def wrapper(self, args, **kwargs):
            location = getattr(args, "location")
            if not location.valid:  # location always must be given
                raise Error("missing repository, please use --repo or BORG_REPO env var!")
            assert isinstance(exclusive, bool)
            lock = getattr(args, "lock", _lock)
            append_only = getattr(args, "append_only", False)
            storage_quota = getattr(args, "storage_quota", None)
            make_parent_dirs = getattr(args, "make_parent_dirs", False)

            repository = get_repository(
                location,
                create=create,
                exclusive=exclusive,
                lock_wait=self.lock_wait,
                lock=lock,
                append_only=append_only,
                make_parent_dirs=make_parent_dirs,
                storage_quota=storage_quota,
                args=args,
            )

            with repository:
                if repository.version not in (2,):
                    raise Error(
                        "This borg version only accepts version 2 repos for -r/--repo. "
                        "You can use 'borg transfer' to copy archives from old to new repos."
                    )
                if manifest or cache:
                    manifest_ = Manifest.load(repository, compatibility)
                    kwargs["manifest"] = manifest_
                    if "compression" in args:
                        manifest_.repo_objs.compressor = args.compression.compressor
                    if secure:
                        assert_secure(repository, manifest_, self.lock_wait)
                if cache:
                    with Cache(
                        repository,
                        manifest_,
                        progress=getattr(args, "progress", False),
                        lock_wait=self.lock_wait,
                        cache_mode=getattr(args, "files_cache_mode", FILES_CACHE_MODE_DISABLED),
                        iec=getattr(args, "iec", False),
                    ) as cache_:
                        return method(self, args, repository=repository, cache=cache_, **kwargs)
                else:
                    return method(self, args, repository=repository, **kwargs)

        return wrapper

    return decorator


def with_other_repository(manifest=False, cache=False, compatibility=None):
    """
    this is a simplified version of "with_repository", just for the "other location".

    the repository at the "other location" is intended to get used as a **source** (== read operations).
    """

    compatibility = compat_check(
        create=False,
        manifest=manifest,
        key=manifest,
        cache=cache,
        compatibility=compatibility,
        decorator_name="with_other_repository",
    )

    def decorator(method):
        @functools.wraps(method)
        def wrapper(self, args, **kwargs):
            location = getattr(args, "other_location")
            if not location.valid:  # nothing to do
                return method(self, args, **kwargs)

            repository = get_repository(
                location,
                create=False,
                exclusive=True,
                lock_wait=self.lock_wait,
                lock=True,
                append_only=False,
                make_parent_dirs=False,
                storage_quota=None,
                args=args,
            )

            with repository:
                if repository.version not in (1, 2):
                    raise Error("This borg version only accepts version 1 or 2 repos for --other-repo.")
                kwargs["other_repository"] = repository
                if manifest or cache:
                    manifest_ = Manifest.load(
                        repository, compatibility, ro_cls=RepoObj if repository.version > 1 else RepoObj1
                    )
                    assert_secure(repository, manifest_, self.lock_wait)
                    if manifest:
                        kwargs["other_manifest"] = manifest_
                if cache:
                    with Cache(
                        repository,
                        manifest_,
                        progress=False,
                        lock_wait=self.lock_wait,
                        cache_mode=getattr(args, "files_cache_mode", FILES_CACHE_MODE_DISABLED),
                        iec=getattr(args, "iec", False),
                    ) as cache_:
                        kwargs["other_cache"] = cache_
                        return method(self, args, **kwargs)
                else:
                    return method(self, args, **kwargs)

        return wrapper

    return decorator


def with_archive(method):
    @functools.wraps(method)
    def wrapper(self, args, repository, manifest, **kwargs):
        archive_name = getattr(args, "name", None)
        assert archive_name is not None
        archive = Archive(
            manifest,
            archive_name,
            numeric_ids=getattr(args, "numeric_ids", False),
            noflags=getattr(args, "noflags", False),
            noacls=getattr(args, "noacls", False),
            noxattrs=getattr(args, "noxattrs", False),
            cache=kwargs.get("cache"),
            log_json=args.log_json,
            iec=args.iec,
        )
        return method(self, args, repository=repository, manifest=manifest, archive=archive, **kwargs)

    return wrapper


class Highlander(argparse.Action):
    """make sure some option is only given once"""

    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest, None) != self.default:
            raise argparse.ArgumentError(self, "There can be only one.")
        setattr(namespace, self.dest, values)


# You can use :ref:`xyz` in the following usage pages. However, for plain-text view,
# e.g. through "borg ... --help", define a substitution for the reference here.
# It will replace the entire :ref:`foo` verbatim.
rst_plain_text_references = {
    "a_status_oddity": '"I am seeing ‘A’ (added) status for a unchanged file!?"',
    "separate_compaction": '"Separate compaction"',
    "list_item_flags": '"Item flags"',
    "borg_patterns": '"borg help patterns"',
    "borg_placeholders": '"borg help placeholders"',
    "key_files": "Internals -> Data structures and file formats -> Key files",
    "borg_key_export": "borg key export --help",
}


def process_epilog(epilog):
    epilog = textwrap.dedent(epilog).splitlines()
    try:
        mode = borg.doc_mode
    except AttributeError:
        mode = "command-line"
    if mode in ("command-line", "build_usage"):
        epilog = [line for line in epilog if not line.startswith(".. man")]
    epilog = "\n".join(epilog)
    if mode == "command-line":
        epilog = rst_to_terminal(epilog, rst_plain_text_references)
    return epilog


def define_exclude_and_patterns(add_option, *, tag_files=False, strip_components=False):
    add_option(
        "-e",
        "--exclude",
        metavar="PATTERN",
        dest="patterns",
        type=parse_exclude_pattern,
        action="append",
        help="exclude paths matching PATTERN",
    )
    add_option(
        "--exclude-from",
        metavar="EXCLUDEFILE",
        action=ArgparseExcludeFileAction,
        help="read exclude patterns from EXCLUDEFILE, one per line",
    )
    add_option(
        "--pattern", metavar="PATTERN", action=ArgparsePatternAction, help="include/exclude paths matching PATTERN"
    )
    add_option(
        "--patterns-from",
        metavar="PATTERNFILE",
        action=ArgparsePatternFileAction,
        help="read include/exclude patterns from PATTERNFILE, one per line",
    )

    if tag_files:
        add_option(
            "--exclude-caches",
            dest="exclude_caches",
            action="store_true",
            help="exclude directories that contain a CACHEDIR.TAG file " "(http://www.bford.info/cachedir/spec.html)",
        )
        add_option(
            "--exclude-if-present",
            metavar="NAME",
            dest="exclude_if_present",
            action="append",
            type=str,
            help="exclude directories that are tagged by containing a filesystem object with " "the given NAME",
        )
        add_option(
            "--keep-exclude-tags",
            dest="keep_exclude_tags",
            action="store_true",
            help="if tag objects are specified with ``--exclude-if-present``, "
            "don't omit the tag objects themselves from the backup archive",
        )

    if strip_components:
        add_option(
            "--strip-components",
            metavar="NUMBER",
            dest="strip_components",
            type=int,
            default=0,
            help="Remove the specified number of leading path elements. "
            "Paths with fewer elements will be silently skipped.",
        )


def define_exclusion_group(subparser, **kwargs):
    exclude_group = subparser.add_argument_group("Exclusion options")
    define_exclude_and_patterns(exclude_group.add_argument, **kwargs)
    return exclude_group


def define_archive_filters_group(subparser, *, sort_by=True, first_last=True, oldest_newest=True, older_newer=True):
    filters_group = subparser.add_argument_group(
        "Archive filters", "Archive filters can be applied to repository targets."
    )
    group = filters_group.add_mutually_exclusive_group()
    group.add_argument(
        "-a",
        "--match-archives",
        metavar="PATTERN",
        dest="match_archives",
        action=Highlander,
        help='only consider archive names matching the pattern. see "borg help match-archives".',
    )

    if sort_by:
        sort_by_default = "timestamp"
        filters_group.add_argument(
            "--sort-by",
            metavar="KEYS",
            dest="sort_by",
            type=SortBySpec,
            default=sort_by_default,
            help="Comma-separated list of sorting keys; valid keys are: {}; default is: {}".format(
                ", ".join(AI_HUMAN_SORT_KEYS), sort_by_default
            ),
        )

    if first_last:
        group = filters_group.add_mutually_exclusive_group()
        group.add_argument(
            "--first",
            metavar="N",
            dest="first",
            default=0,
            type=positive_int_validator,
            help="consider first N archives after other filters were applied",
        )
        group.add_argument(
            "--last",
            metavar="N",
            dest="last",
            default=0,
            type=positive_int_validator,
            help="consider last N archives after other filters were applied",
        )

    if oldest_newest:
        group = filters_group.add_mutually_exclusive_group()
        group.add_argument(
            "--oldest",
            metavar="TIMESPAN",
            type=relative_time_marker_validator,
            dest="oldest",
            help="consider archives between the oldest archive's timestamp and (oldest + TIMESPAN), e.g. 7d or 12m.",
        )
        group.add_argument(
            "--newest",
            metavar="TIMESPAN",
            type=relative_time_marker_validator,
            dest="newest",
            help="consider archives between the newest archive's timestamp and (newest - TIMESPAN), e.g. 7d or 12m.",
        )

    if older_newer:
        group = filters_group.add_mutually_exclusive_group()
        group.add_argument(
            "--older",
            metavar="TIMESPAN",
            type=relative_time_marker_validator,
            dest="older",
            help="consider archives older than (now - TIMESPAN), e.g. 7d oder 12m.",
        )
        group.add_argument(
            "--newer",
            metavar="TIMESPAN",
            type=relative_time_marker_validator,
            dest="newer",
            help="consider archives newer than (now - TIMESPAN), e.g. 7d or 12m.",
        )

    return filters_group


def define_common_options(add_common_option):
    add_common_option("-h", "--help", action="help", help="show this help message and exit")
    add_common_option(
        "--critical",
        dest="log_level",
        action="store_const",
        const="critical",
        default="warning",
        help="work on log level CRITICAL",
    )
    add_common_option(
        "--error",
        dest="log_level",
        action="store_const",
        const="error",
        default="warning",
        help="work on log level ERROR",
    )
    add_common_option(
        "--warning",
        dest="log_level",
        action="store_const",
        const="warning",
        default="warning",
        help="work on log level WARNING (default)",
    )
    add_common_option(
        "--info",
        "-v",
        "--verbose",
        dest="log_level",
        action="store_const",
        const="info",
        default="warning",
        help="work on log level INFO",
    )
    add_common_option(
        "--debug",
        dest="log_level",
        action="store_const",
        const="debug",
        default="warning",
        help="enable debug output, work on log level DEBUG",
    )
    add_common_option(
        "--debug-topic",
        metavar="TOPIC",
        dest="debug_topics",
        action="append",
        default=[],
        help="enable TOPIC debugging (can be specified multiple times). "
        "The logger path is borg.debug.<TOPIC> if TOPIC is not fully qualified.",
    )
    add_common_option("-p", "--progress", dest="progress", action="store_true", help="show progress information")
    add_common_option("--iec", dest="iec", action="store_true", help="format using IEC units (1KiB = 1024B)")
    add_common_option(
        "--log-json",
        dest="log_json",
        action="store_true",
        help="Output one JSON object per log line instead of formatted text.",
    )
    add_common_option(
        "--lock-wait",
        metavar="SECONDS",
        dest="lock_wait",
        type=int,
        default=int(os.environ.get("BORG_LOCK_WAIT", 1)),
        help="wait at most SECONDS for acquiring a repository/cache lock (default: %(default)d).",
    )
    add_common_option(
        "--bypass-lock",
        dest="lock",
        action="store_false",
        default=argparse.SUPPRESS,  # only create args attribute if option is specified
        help="Bypass locking mechanism",
    )
    add_common_option("--show-version", dest="show_version", action="store_true", help="show/log the borg version")
    add_common_option("--show-rc", dest="show_rc", action="store_true", help="show/log the return code (rc)")
    add_common_option(
        "--umask",
        metavar="M",
        dest="umask",
        type=lambda s: int(s, 8),
        default=UMASK_DEFAULT,
        help="set umask to M (local only, default: %(default)04o)",
    )
    add_common_option(
        "--remote-path",
        metavar="PATH",
        dest="remote_path",
        help='use PATH as borg executable on the remote (default: "borg")',
    )
    add_common_option(
        "--upload-ratelimit",
        metavar="RATE",
        dest="upload_ratelimit",
        type=int,
        help="set network upload rate limit in kiByte/s (default: 0=unlimited)",
    )
    add_common_option(
        "--upload-buffer",
        metavar="UPLOAD_BUFFER",
        dest="upload_buffer",
        type=int,
        help="set network upload buffer size in MiB. (default: 0=no buffer)",
    )
    add_common_option(
        "--debug-profile",
        metavar="FILE",
        dest="debug_profile",
        default=None,
        help="Write execution profile in Borg format into FILE. For local use a Python-"
        'compatible file can be generated by suffixing FILE with ".pyprof".',
    )
    add_common_option(
        "--rsh",
        metavar="RSH",
        dest="rsh",
        help="Use this command to connect to the 'borg serve' process (default: 'ssh')",
    )
    add_common_option(
        "-r",
        "--repo",
        metavar="REPO",
        dest="location",
        type=location_validator(other=False),
        default=Location(other=False),
        help="repository to use",
    )


def build_matcher(inclexcl_patterns, include_paths):
    matcher = PatternMatcher()
    matcher.add_inclexcl(inclexcl_patterns)
    matcher.add_includepaths(include_paths)
    return matcher


def build_filter(matcher, strip_components):
    if strip_components:

        def item_filter(item):
            matched = matcher.match(item.path) and os.sep.join(item.path.split(os.sep)[strip_components:])
            return matched

    else:

        def item_filter(item):
            matched = matcher.match(item.path)
            return matched

    return item_filter



from ...constants import *  # NOQA
from . import cmd, changedir


def test_return_codes(cmd, tmpdir):
    repo = tmpdir.mkdir("repo")
    input = tmpdir.mkdir("input")
    output = tmpdir.mkdir("output")
    input.join("test_file").write("content")
    rc, out = cmd("--repo=%s" % str(repo), "rcreate", "--encryption=none")
    assert rc == EXIT_SUCCESS
    rc, out = cmd("--repo=%s" % repo, "create", "archive", str(input))
    assert rc == EXIT_SUCCESS
    with changedir(str(output)):
        rc, out = cmd("--repo=%s" % repo, "extract", "archive")
        assert rc == EXIT_SUCCESS
    rc, out = cmd("--repo=%s" % repo, "extract", "archive", "does/not/match")
    assert rc == EXIT_WARNING  # pattern did not match
    rc, out = cmd("--repo=%s" % repo, "create", "archive", str(input))
    assert rc == EXIT_ERROR  # duplicate archive name



# Note: these tests are part of the self test, do not use or import pytest functionality here.
#       See borg.selftest for details. If you add/remove test methods, update SELFTEST_COUNT

from binascii import hexlify
from unittest.mock import MagicMock
import unittest

from ..crypto.low_level import AES256_CTR_HMAC_SHA256, AES256_OCB, CHACHA20_POLY1305, UNENCRYPTED, IntegrityError
from ..crypto.low_level import bytes_to_long, bytes_to_int, long_to_bytes
from ..crypto.low_level import hkdf_hmac_sha512
from ..crypto.low_level import AES, hmac_sha256
from ..crypto.key import CHPOKeyfileKey, AESOCBRepoKey, FlexiKey
from ..helpers import msgpack

from . import BaseTestCase


class CryptoTestCase(BaseTestCase):
    def test_bytes_to_int(self):
        self.assert_equal(bytes_to_int(b"\0\0\0\1"), 1)

    def test_bytes_to_long(self):
        self.assert_equal(bytes_to_long(b"\0\0\0\0\0\0\0\1"), 1)
        self.assert_equal(long_to_bytes(1), b"\0\0\0\0\0\0\0\1")

    def test_UNENCRYPTED(self):
        iv = b""  # any IV is ok, it just must be set and not None
        data = b"data"
        header = b"header"
        cs = UNENCRYPTED(None, None, iv, header_len=6)
        envelope = cs.encrypt(data, header=header)
        self.assert_equal(envelope, header + data)
        got_data = cs.decrypt(envelope)
        self.assert_equal(got_data, data)

    def test_AES256_CTR_HMAC_SHA256(self):
        # this tests the layout as in borg < 1.2 (1 type byte, no aad)
        mac_key = b"Y" * 32
        enc_key = b"X" * 32
        iv = 0
        data = b"foo" * 10
        header = b"\x42"
        # encrypt-then-mac
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, iv, header_len=1, aad_offset=1)
        hdr_mac_iv_cdata = cs.encrypt(data, header=header)
        hdr = hdr_mac_iv_cdata[0:1]
        mac = hdr_mac_iv_cdata[1:33]
        iv = hdr_mac_iv_cdata[33:41]
        cdata = hdr_mac_iv_cdata[41:]
        self.assert_equal(hexlify(hdr), b"42")
        self.assert_equal(hexlify(mac), b"af90b488b0cc4a8f768fe2d6814fa65aec66b148135e54f7d4d29a27f22f57a8")
        self.assert_equal(hexlify(iv), b"0000000000000000")
        self.assert_equal(hexlify(cdata), b"c6efb702de12498f34a2c2bbc8149e759996d08bf6dc5c610aefc0c3a466")
        self.assert_equal(cs.next_iv(), 2)
        # auth-then-decrypt
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, header_len=len(header), aad_offset=1)
        pdata = cs.decrypt(hdr_mac_iv_cdata)
        self.assert_equal(data, pdata)
        self.assert_equal(cs.next_iv(), 2)
        # auth-failure due to corruption (corrupted data)
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, header_len=len(header), aad_offset=1)
        hdr_mac_iv_cdata_corrupted = hdr_mac_iv_cdata[:41] + b"\0" + hdr_mac_iv_cdata[42:]
        self.assert_raises(IntegrityError, lambda: cs.decrypt(hdr_mac_iv_cdata_corrupted))

    def test_AES256_CTR_HMAC_SHA256_aad(self):
        mac_key = b"Y" * 32
        enc_key = b"X" * 32
        iv = 0
        data = b"foo" * 10
        header = b"\x12\x34\x56"
        # encrypt-then-mac
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, iv, header_len=3, aad_offset=1)
        hdr_mac_iv_cdata = cs.encrypt(data, header=header)
        hdr = hdr_mac_iv_cdata[0:3]
        mac = hdr_mac_iv_cdata[3:35]
        iv = hdr_mac_iv_cdata[35:43]
        cdata = hdr_mac_iv_cdata[43:]
        self.assert_equal(hexlify(hdr), b"123456")
        self.assert_equal(hexlify(mac), b"7659a915d9927072ef130258052351a17ef882692893c3850dd798c03d2dd138")
        self.assert_equal(hexlify(iv), b"0000000000000000")
        self.assert_equal(hexlify(cdata), b"c6efb702de12498f34a2c2bbc8149e759996d08bf6dc5c610aefc0c3a466")
        self.assert_equal(cs.next_iv(), 2)
        # auth-then-decrypt
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, header_len=len(header), aad_offset=1)
        pdata = cs.decrypt(hdr_mac_iv_cdata)
        self.assert_equal(data, pdata)
        self.assert_equal(cs.next_iv(), 2)
        # auth-failure due to corruption (corrupted aad)
        cs = AES256_CTR_HMAC_SHA256(mac_key, enc_key, header_len=len(header), aad_offset=1)
        hdr_mac_iv_cdata_corrupted = hdr_mac_iv_cdata[:1] + b"\0" + hdr_mac_iv_cdata[2:]
        self.assert_raises(IntegrityError, lambda: cs.decrypt(hdr_mac_iv_cdata_corrupted))

    def test_AE(self):
        # used in legacy-like layout (1 type byte, no aad)
        key = b"X" * 32
        iv_int = 0
        data = b"foo" * 10
        header = b"\x23" + iv_int.to_bytes(12, "big")
        tests = [
            # (ciphersuite class, exp_mac, exp_cdata)
            (
                AES256_OCB,
                b"b6909c23c9aaebd9abbe1ff42097652d",
                b"877ce46d2f62dee54699cebc3ba41d9ab613f7c486778c1b3636664b1493",
            ),
            (
                CHACHA20_POLY1305,
                b"fd08594796e0706cde1e8b461e3e0555",
                b"a093e4b0387526f085d3c40cca84a35230a5c0dd766453b77ba38bcff775",
            ),
        ]
        for cs_cls, exp_mac, exp_cdata in tests:
            # print(repr(cs_cls))
            # encrypt/mac
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            hdr_mac_iv_cdata = cs.encrypt(data, header=header)
            hdr = hdr_mac_iv_cdata[0:1]
            iv = hdr_mac_iv_cdata[1:13]
            mac = hdr_mac_iv_cdata[13:29]
            cdata = hdr_mac_iv_cdata[29:]
            self.assert_equal(hexlify(hdr), b"23")
            self.assert_equal(hexlify(mac), exp_mac)
            self.assert_equal(hexlify(iv), b"000000000000000000000000")
            self.assert_equal(hexlify(cdata), exp_cdata)
            self.assert_equal(cs.next_iv(), 1)
            # auth/decrypt
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            pdata = cs.decrypt(hdr_mac_iv_cdata)
            self.assert_equal(data, pdata)
            self.assert_equal(cs.next_iv(), 1)
            # auth-failure due to corruption (corrupted data)
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            hdr_mac_iv_cdata_corrupted = hdr_mac_iv_cdata[:29] + b"\0" + hdr_mac_iv_cdata[30:]
            self.assert_raises(IntegrityError, lambda: cs.decrypt(hdr_mac_iv_cdata_corrupted))

    def test_AEAD(self):
        # test with aad
        key = b"X" * 32
        iv_int = 0
        data = b"foo" * 10
        header = b"\x12\x34\x56" + iv_int.to_bytes(12, "big")
        tests = [
            # (ciphersuite class, exp_mac, exp_cdata)
            (
                AES256_OCB,
                b"f2748c412af1c7ead81863a18c2c1893",
                b"877ce46d2f62dee54699cebc3ba41d9ab613f7c486778c1b3636664b1493",
            ),
            (
                CHACHA20_POLY1305,
                b"b7e7c9a79f2404e14f9aad156bf091dd",
                b"a093e4b0387526f085d3c40cca84a35230a5c0dd766453b77ba38bcff775",
            ),
        ]
        for cs_cls, exp_mac, exp_cdata in tests:
            # print(repr(cs_cls))
            # encrypt/mac
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            hdr_mac_iv_cdata = cs.encrypt(data, header=header)
            hdr = hdr_mac_iv_cdata[0:3]
            iv = hdr_mac_iv_cdata[3:15]
            mac = hdr_mac_iv_cdata[15:31]
            cdata = hdr_mac_iv_cdata[31:]
            self.assert_equal(hexlify(hdr), b"123456")
            self.assert_equal(hexlify(mac), exp_mac)
            self.assert_equal(hexlify(iv), b"000000000000000000000000")
            self.assert_equal(hexlify(cdata), exp_cdata)
            self.assert_equal(cs.next_iv(), 1)
            # auth/decrypt
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            pdata = cs.decrypt(hdr_mac_iv_cdata)
            self.assert_equal(data, pdata)
            self.assert_equal(cs.next_iv(), 1)
            # auth-failure due to corruption (corrupted aad)
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=1)
            hdr_mac_iv_cdata_corrupted = hdr_mac_iv_cdata[:1] + b"\0" + hdr_mac_iv_cdata[2:]
            self.assert_raises(IntegrityError, lambda: cs.decrypt(hdr_mac_iv_cdata_corrupted))

    def test_AEAD_with_more_AAD(self):
        # test giving extra aad to the .encrypt() and .decrypt() calls
        key = b"X" * 32
        iv_int = 0
        data = b"foo" * 10
        header = b"\x12\x34"
        tests = [AES256_OCB, CHACHA20_POLY1305]
        for cs_cls in tests:
            # encrypt/mac
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=0)
            hdr_mac_iv_cdata = cs.encrypt(data, header=header, aad=b"correct_chunkid")
            # successful auth/decrypt (correct aad)
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=0)
            pdata = cs.decrypt(hdr_mac_iv_cdata, aad=b"correct_chunkid")
            self.assert_equal(data, pdata)
            # unsuccessful auth (incorrect aad)
            cs = cs_cls(key, iv_int, header_len=len(header), aad_offset=0)
            self.assert_raises(IntegrityError, lambda: cs.decrypt(hdr_mac_iv_cdata, aad=b"incorrect_chunkid"))

    # These test vectors come from https://www.kullo.net/blog/hkdf-sha-512-test-vectors/
    # who claims to have verified these against independent Python and C++ implementations.

    def test_hkdf_hmac_sha512(self):
        ikm = b"\x0b" * 22
        salt = bytes.fromhex("000102030405060708090a0b0c")
        info = bytes.fromhex("f0f1f2f3f4f5f6f7f8f9")
        l = 42

        okm = hkdf_hmac_sha512(ikm, salt, info, l)
        assert okm == bytes.fromhex(
            "832390086cda71fb47625bb5ceb168e4c8e26a1a16ed34d9fc7fe92c1481579338da362cb8d9f925d7cb"
        )

    def test_hkdf_hmac_sha512_2(self):
        ikm = bytes.fromhex(
            "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f2021222324252627"
            "28292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f404142434445464748494a4b4c4d4e4f"
        )
        salt = bytes.fromhex(
            "606162636465666768696a6b6c6d6e6f707172737475767778797a7b7c7d7e7f808182838485868"
            "788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeaf"
        )
        info = bytes.fromhex(
            "b0b1b2b3b4b5b6b7b8b9babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecfd0d1d2d3d4d5d6d7"
            "d8d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7f8f9fafbfcfdfeff"
        )
        l = 82

        okm = hkdf_hmac_sha512(ikm, salt, info, l)
        assert okm == bytes.fromhex(
            "ce6c97192805b346e6161e821ed165673b84f400a2b514b2fe23d84cd189ddf1b695b48cbd1c838844"
            "1137b3ce28f16aa64ba33ba466b24df6cfcb021ecff235f6a2056ce3af1de44d572097a8505d9e7a93"
        )

    def test_hkdf_hmac_sha512_3(self):
        ikm = bytes.fromhex("0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b")
        salt = None
        info = b""
        l = 42

        okm = hkdf_hmac_sha512(ikm, salt, info, l)
        assert okm == bytes.fromhex(
            "f5fa02b18298a72a8c23898a8703472c6eb179dc204c03425c970e3b164bf90fff22d04836d0e2343bac"
        )

    def test_hkdf_hmac_sha512_4(self):
        ikm = bytes.fromhex("0b0b0b0b0b0b0b0b0b0b0b")
        salt = bytes.fromhex("000102030405060708090a0b0c")
        info = bytes.fromhex("f0f1f2f3f4f5f6f7f8f9")
        l = 42

        okm = hkdf_hmac_sha512(ikm, salt, info, l)
        assert okm == bytes.fromhex(
            "7413e8997e020610fbf6823f2ce14bff01875db1ca55f68cfcf3954dc8aff53559bd5e3028b080f7c068"
        )

    def test_hkdf_hmac_sha512_5(self):
        ikm = bytes.fromhex("0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c")
        salt = None
        info = b""
        l = 42

        okm = hkdf_hmac_sha512(ikm, salt, info, l)
        assert okm == bytes.fromhex(
            "1407d46013d98bc6decefcfee55f0f90b0c7f63d68eb1a80eaf07e953cfc0a3a5240a155d6e4daa965bb"
        )


def test_decrypt_key_file_argon2_chacha20_poly1305():
    plain = b"hello"
    # echo -n "hello, pass phrase" | argon2 saltsaltsaltsalt -id -t 1 -k 8 -p 1 -l 32 -r
    key = bytes.fromhex("a1b0cba145c154fbd8960996c5ce3428e9920cfe53c84ef08b4102a70832bcec")
    ae_cipher = CHACHA20_POLY1305(key=key, iv=0, header_len=0, aad_offset=0)

    envelope = ae_cipher.encrypt(plain)

    encrypted = msgpack.packb(
        {
            "version": 1,
            "salt": b"salt" * 4,
            "argon2_time_cost": 1,
            "argon2_memory_cost": 8,
            "argon2_parallelism": 1,
            "argon2_type": b"id",
            "algorithm": "argon2 chacha20-poly1305",
            "data": envelope,
        }
    )
    key = CHPOKeyfileKey(None)

    decrypted = key.decrypt_key_file(encrypted, "hello, pass phrase")

    assert decrypted == plain


def test_decrypt_key_file_pbkdf2_sha256_aes256_ctr_hmac_sha256():
    plain = b"hello"
    salt = b"salt" * 4
    passphrase = "hello, pass phrase"
    key = FlexiKey.pbkdf2(passphrase, salt, 1, 32)
    hash = hmac_sha256(key, plain)
    data = AES(key, b"\0" * 16).encrypt(plain)
    encrypted = msgpack.packb(
        {"version": 1, "algorithm": "sha256", "iterations": 1, "salt": salt, "data": data, "hash": hash}
    )
    key = CHPOKeyfileKey(None)

    decrypted = key.decrypt_key_file(encrypted, passphrase)

    assert decrypted == plain


@unittest.mock.patch("getpass.getpass")
def test_repo_key_detect_does_not_raise_integrity_error(getpass, monkeypatch):
    """https://github.com/borgbackup/borg/pull/6469#discussion_r832670411

    This is a regression test for a bug I introduced and fixed:

    Traceback (most recent call last):
      File "/home/user/borg-master/src/borg/testsuite/crypto.py", line 384, in test_repo_key_detect_does_not_raise_integrity_error
        RepoKey.detect(repository, manifest_data=None)
      File "/home/user/borg-master/src/borg/crypto/key.py", line 402, in detect
        if not key.load(target, passphrase):
      File "/home/user/borg-master/src/borg/crypto/key.py", line 654, in load
        success = self._load(key_data, passphrase)
      File "/home/user/borg-master/src/borg/crypto/key.py", line 418, in _load
        data = self.decrypt_key_file(cdata, passphrase)
      File "/home/user/borg-master/src/borg/crypto/key.py", line 444, in decrypt_key_file
        return self.decrypt_key_file_argon2(encrypted_key, passphrase)
      File "/home/user/borg-master/src/borg/crypto/key.py", line 470, in decrypt_key_file_argon2
        return ae_cipher.decrypt(encrypted_key.data)
      File "src/borg/crypto/low_level.pyx", line 302, in borg.crypto.low_level.AES256_CTR_BASE.decrypt
        self.mac_verify(<const unsigned char *> idata.buf+aoffset, alen,
      File "src/borg/crypto/low_level.pyx", line 382, in borg.crypto.low_level.AES256_CTR_HMAC_SHA256.mac_verify
        raise IntegrityError('MAC Authentication failed')
    borg.crypto.low_level.IntegrityError: MAC Authentication failed

    1. FlexiKey.decrypt_key_file() is supposed to signal the decryption failure by returning None
    2. FlexiKey.detect() relies on that interface - it tries an empty passphrase before prompting the user
    3. my initial implementation of decrypt_key_file_argon2() was simply passing through the IntegrityError() from AES256_CTR_BASE.decrypt()
    """
    repository = MagicMock(id=b"repository_id")
    getpass.return_value = "hello, pass phrase"
    monkeypatch.setenv("BORG_DISPLAY_PASSPHRASE", "no")
    AESOCBRepoKey.create(repository, args=MagicMock(key_algorithm="argon2"))
    repository.load_key.return_value = repository.save_key.call_args.args[0]

    AESOCBRepoKey.detect(repository, manifest_data=None)
